

import math

from decimal import Decimal

from .. import tag

from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class Arrow1(ClosedWidget):
    """An svg arrow shape, fitting in a 100x100 space
    """

    # This class does not display any error messages
    display_errors = False

    _points = ((49,1), (50,1), (98,30), (98,32), (60,32), (60,98), (39,98), (39,32), (1,32), (1,30))

    arg_descriptions = {'fill':FieldArg("text", "none", jsonset=True),
                        'stroke':FieldArg("text", "black", jsonset=True),
                        'transform':FieldArg("text", "", jsonset=True),
                        'stroke_width':FieldArg("text", "1", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        fill: The fill colour, use none for no fill
        stroke: The outline edge colour
        stroke_width: The outline edge thickness
        transform: The svg transform object, use it to scals and rotate
        """
        ClosedWidget.__init__(self, name=name, tag_name="polygon", brief=brief, **field_args)

    def _build(self, page, ident_list, environ, call_data, lang):
        "create arrow"                  
        attrib_points = ""
        for p in self._points:
            point = "%s, %s " % p
            attrib_points += point
        self.update_attribs({"points":attrib_points})

        if self.get_field_value("fill"):
            self.update_attribs({"fill":self.get_field_value("fill")})
        if self.get_field_value("stroke"):
            self.update_attribs({"stroke":self.get_field_value("stroke")})
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})
        if self.get_field_value("stroke_width"):
            self.update_attribs({"stroke-width":self.get_field_value("stroke_width")})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<polygon /> <!-- arrow shape with widget id, class widget_class and the given attributes -->"""



class Arrow2(ClosedWidget):
    """A slim svg arrow shape, fitting in a 50x200 space
    """

    # This class does not display any error messages
    display_errors = False

    _points = ((24,1), (25,1), (49,50), (49,52), (30,52), (30,198), (19,198), (19,52), (1,52), (1,50))

    arg_descriptions = {'fill':FieldArg("text", "none", jsonset=True),
                        'stroke':FieldArg("text", "black", jsonset=True),
                        'transform':FieldArg("text", "", jsonset=True),
                        'stroke_width':FieldArg("text", "1", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        fill: The fill colour, use none for no fill
        stroke: The outline edge colour
        stroke_width: The outline edge thickness
        transform: The svg transform object, use it to scals and rotate
        """
        ClosedWidget.__init__(self, name=name, tag_name="polygon", brief=brief, **field_args)

    def _build(self, page, ident_list, environ, call_data, lang):
        "create arrow"                  
        attrib_points = ""
        for p in self._points:
            point = "%s, %s " % p
            attrib_points += point
        self.update_attribs({"points":attrib_points})

        if self.get_field_value("fill"):
            self.update_attribs({"fill":self.get_field_value("fill")})
        if self.get_field_value("stroke"):
            self.update_attribs({"stroke":self.get_field_value("stroke")})
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})
        if self.get_field_value("stroke_width"):
            self.update_attribs({"stroke-width":self.get_field_value("stroke_width")})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<polygon /> <!-- arrow shape with widget id, class widget_class and the given attributes -->"""


class Vertical1(Widget):

    # This class does not display any error messages
    display_errors = False

    _points = ((110,49), (110,50), (81,98), (79,98), (79,60), (13,60), (13,39), (79,39), (79,1), (81,1))

    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True),
                        'arrow_fill':FieldArg("text", "blue", jsonset=True),
                        'minimum':FieldArg("text", "0"),
                        'maximum':FieldArg("text", "100"),
                        'smallintervals':FieldArg("text", "10"),
                        'largeintervals':FieldArg("text", "20"),
                        'measurement':FieldArg("text", "50", jsonset=True),
                        'font_family':FieldArg("text", "arial")
                       }


    def _make_scale(self, minimum, maximum, smallintervals, largeintervals):
        "Returns two lists of Decimal values"

        minvalue = Decimal(minimum)
        maxvalue = Decimal(maximum)
        smallint = Decimal(smallintervals)
        largeint = Decimal(largeintervals)
       
        # start at the bottom of the scale with minvalue
        minscale = [minvalue]
        maxscale = [minvalue]
        mns = minvalue
        mxs = minvalue

        while mxs < maxvalue:
            mxs += largeint
            maxscale.append(mxs)

        while True:
            mns += smallint
            if mns > maxscale[-1]:
                break
            minscale.append(mns)
        return minscale, maxscale

    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds a vertical scale and arrow, held in a 700 high x 250 wide space"""
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self[0] = tag.ClosedPart(tag_name='rect', attribs={"x":"100",
                                                           "y":"1",
                                                           "rx":"2",
                                                           "ry":"2",
                                                           "width":"149",
                                                           "height":"698",
                                                           "fill":"white",
                                                           "stroke":"black",
                                                           "stroke-width":"1"})
        arrow_points = ""
        for p in self._points:
            point = "%s, %s " % p
            arrow_points += point
        self[1] = tag.ClosedPart(tag_name='polygon', attribs={
                                                           "fill":"white",
                                                           "stroke":"black",
                                                           "stroke-width":"2",
                                                           "points":arrow_points })
        self[2] = tag.ClosedPart(tag_name='line', attribs={
                                                            'x1':'120',
                                                            'y1':'50',
                                                            'x2':'120',
                                                            'y2':'650',
                                                            'stroke':"black",
                                                            'stroke-width':"2"  })


    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

        font_family = self.get_field_value("font_family")
        if not font_family:
            font_family = "arial"

        if self.get_field_value("arrow_fill"):
            self[1].update_attribs({"fill":self.get_field_value("arrow_fill")})
        # make the scale
        minscale, maxscale = self._make_scale(self.get_field_value("minimum"),
                                              self.get_field_value("maximum"),
                                              self.get_field_value("smallintervals"),
                                              self.get_field_value("largeintervals"))

        # small lines
        minitems = len(minscale)
        scalemins = Decimal('600.0') / (minitems-1)
        n = 3
        for index, item in enumerate(minscale):
            vert = Decimal(650) - index*scalemins
            self[n] = tag.ClosedPart(tag_name='line', attribs={
                                                            'x1':'120',
                                                            'y1': str(vert),
                                                            'x2':'150',
                                                            'y2':str(vert),
                                                            'stroke':"black",
                                                            'stroke-width':"1"  })
            n += 1

        # large lines
        maxitems = len(maxscale)
        scalemaxs = Decimal('600.0') / (maxitems-1)
        for index, item in enumerate(maxscale):
            vert = Decimal('650') - index*scalemaxs
            self[n] = tag.ClosedPart(tag_name='line', attribs={
                                                            'x1':'119',
                                                            'y1': str(vert),
                                                            'x2':'210',
                                                            'y2':str(vert),
                                                            'stroke':"black",
                                                            'stroke-width':"3"  })

            n += 1
            self[n] = tag.Part(tag_name='text', text=str(item), attribs={
                                                            'x':'160',
                                                            'y': str(vert-10),
                                                            'font-size': '20',
                                                            'font-family': font_family,
                                                            'stroke':"black",
                                                            'stroke-width':"1"  })

            n += 1

        # now place arrow at the measurement point
        measurement = Decimal(self.get_field_value("measurement"))
        self._minvalue = maxscale[0]
        self._maxvalue = maxscale[-1]
        if measurement >= self._maxvalue:
            return
        if measurement <= self._minvalue:
            self[1].update_attribs({"transform" : "translate(0, 600)"})
            return
        m = Decimal('600.0') - (measurement - self._minvalue)*600/(self._maxvalue-self._minvalue)
        self[1].update_attribs({"transform" : "translate(0, %s)" % (m,)})

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sends scaling factor for mapping measurement to scale"""
        return self._make_fieldvalues(maxvalue=str(self._maxvalue), minvalue=str(self._minvalue))


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
  <rect /> <!-- the scale rectangle -->
  <!-- lines and text dependent on the input scale values -->
  <polygon /> <!-- the arrow, with translate linked to the input value -->
</g>"""




class Traditional1(Widget):

    # This class does not display any error messages
    display_errors = False

    _points = ((24,1), (25,1), (49,50), (49,52), (30,52), (30,198), (19,198), (19,52), (1,52), (1,50))

    arg_descriptions = {
                        'transform':FieldArg("text", "translate(10,10)", jsonset=True),
                        'minimum':FieldArg("text", "0"),
                        'maximum':FieldArg("text", "100"),
                        'smallintervals':FieldArg("text", "10"),
                        'largeintervals':FieldArg("text", "20"),
                        'arrow_stroke':FieldArg("text", "grey", jsonset=True),
                        'measurement':FieldArg("text", "50", jsonset=True),
                        'font_family':FieldArg("text", "arial")
                       }


    def _make_scale(self, minimum, maximum, smallintervals, largeintervals):
        "Returns two lists of Decimal values"

        minvalue = Decimal(minimum)
        maxvalue = Decimal(maximum)
        smallint = Decimal(smallintervals)
        largeint = Decimal(largeintervals)
       
        # start at the bottom of the scale with minvalue
        minscale = [minvalue]
        maxscale = [minvalue]
        mns = minvalue
        mxs = minvalue

        while mxs < maxvalue:
            mxs += largeint
            maxscale.append(mxs)

        while True:
            mns += smallint
            if mns > maxscale[-1]:
                break
            minscale.append(mns)
        return minscale, maxscale

    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds the scale and arrow, held in a 400 high x 700 wide space"""
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)

        # A path which holds the curved shape which will contain the meter

        # the angle of the white backing is 140 degrees, this makes an
        # angle of 20 degrees to the horizonta. So get this in radians
        back_horizontal_angle = math.radians(20.0)

        # The scale

        # the angle of the scale, 120 degrees
        scale_angle = 120
        self._scale_angle = Decimal(scale_angle)

        # the angle to the horizontal, 30 degrees, get it in radians
        scale_horizontal_angle = math.radians((180-scale_angle)/2.0)

        # radius of outside of white backing shape
        r1 = 320

        # radius of scale line
        r2 = 230
        self._scale_r = r2

        # radius of inside of white backing shape
        r3 = 200

        # coordinates of rotation centre of the meter
        cx = 350
        cy = 350
        self._cx = cx
        self._cy = cy

        # create white backing shape
        left_out_x = cx - r1*math.cos(back_horizontal_angle)
        left_out_y = cy - r1*math.sin(back_horizontal_angle)

        right_out_x = cx + r1*math.cos(back_horizontal_angle)
        right_out_y = left_out_y

        right_in_x = cx + r3*math.cos(back_horizontal_angle)
        right_in_y = cy - r3*math.sin(back_horizontal_angle)

        left_in_x = cx - r3*math.cos(back_horizontal_angle)
        left_in_y = right_in_y

        path_data = """
M %s %s
A %s %s 0 0 1 %s %s
L %s %s
A %s %s 0 0 0 %s %s
Z""" % (left_out_x, left_out_y,
        r1, r1, right_out_x, right_out_y,
        right_in_x, right_in_y,
        r3, r3, left_in_x, left_in_y)

        self[0] = tag.ClosedPart(tag_name='path',
                                 attribs={"fill":"white", "stroke":"black", "stroke-width":"1", "d":path_data})

        # create the scale curve

        # still centred on cx, cy
        scale_left_x = cx - r2 * math.cos(scale_horizontal_angle)
        scale_left_y = cy - r2 * math.sin(scale_horizontal_angle)
        scale_right_x = cx + r2 * math.cos(scale_horizontal_angle)
        scale_right_y = scale_left_y

        # Draw the scale curve
        scale_data = """
M %s %s
A %s %s 0 0 1 %s %s
""" % (scale_left_x, scale_left_y,
       r2, r2, scale_right_x, scale_right_y,)

        self[1] = tag.ClosedPart(tag_name='path',
                                 attribs={ "fill":"none", "stroke":"black", "stroke-width":"2", "d":scale_data})


        # The arrow points
        arrow_points = ""
        # move all points to the right and down,
        # note 24.5 is x distance to arrow point
        x_move = cx - 24.5
        # moves arrow down to just touch the scale
        y_move = cy - r2
        for p in self._points:
            point = "%s, %s " % (p[0] + x_move, p[1] + y_move)
            arrow_points += point
        self[2] = tag.ClosedPart(tag_name='polygon', attribs={
                                                           "fill":"black",
                                                           "stroke":"grey",
                                                           "stroke-width":"2",
                                                           "points":arrow_points })

        # insert a circle at arrow hub, of radius 40
        self[3] = tag.ClosedPart(tag_name='circle', attribs={
                                                           "cx": str(cx),
                                                           "cy": str(cy),
                                                           "r": "40",
                                                           "fill":"black",
                                                           "stroke":"grey",
                                                           "stroke-width":"2" })



    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

        font_family = self.get_field_value("font_family")
        if not font_family:
            font_family = "arial"

        if self.get_field_value("arrow_stroke"):
            self[2].update_attribs({"stroke":self.get_field_value("arrow_stroke")})
            self[3].update_attribs({"stroke":self.get_field_value("arrow_stroke")})
        # make the scale
        minscale, maxscale = self._make_scale(self.get_field_value("minimum"),
                                              self.get_field_value("maximum"),
                                              self.get_field_value("smallintervals"),
                                              self.get_field_value("largeintervals"))

        # start angle is 180 - 120 / 2 normally 30
        start_angle = (Decimal('180') - self._scale_angle)/Decimal('2')

        # small lines, each of length 20
        minitems = len(minscale)
        scalemindegs = self._scale_angle / (minitems-1)
        line_r = self._scale_r + 20
        n = 4
        for index, item in enumerate(minscale):
            angle = start_angle + index*scalemindegs
            rads = math.radians(float(angle))

            x1 = self._cx - self._scale_r * math.cos(rads)
            y1 = self._cy - self._scale_r * math.sin(rads)
            x2 = self._cx - line_r * math.cos(rads)
            y2 = self._cy - line_r * math.sin(rads)

            self[n] = tag.ClosedPart(tag_name='line', attribs={
                                                            'x1':str(x1),
                                                            'y1':str(y1),
                                                            'x2':str(x2),
                                                            'y2':str(y2),
                                                            'stroke':"black",
                                                            'stroke-width':"1"  })
            n += 1

        # large lines, each of length 40
        maxitems = len(maxscale)
        scalemaxdegs = self._scale_angle / (maxitems-1)
        line_r = self._scale_r + 40
        # slightly shorter r as the curved line has stroke width of 2
        reduced_r = self._scale_r - 1
        for index, item in enumerate(maxscale):
            angle = start_angle + index*scalemaxdegs
            rads = math.radians(float(angle))

            x1 = self._cx - reduced_r * math.cos(rads)
            y1 = self._cy - reduced_r * math.sin(rads)
            x2 = self._cx - line_r * math.cos(rads)
            y2 = self._cy - line_r * math.sin(rads)

            self[n] = tag.ClosedPart(tag_name='line', attribs={
                                                            'x1':str(x1),
                                                            'y1':str(y1),
                                                            'x2':str(x2),
                                                            'y2':str(y2),
                                                            'stroke':"black",
                                                            'stroke-width':"3"  })
            n += 1
            self[n] = tag.Part(tag_name='text', text=str(item), attribs={
                                                            'x':str(x2-10),
                                                            'y': str(y2-5),
                                                            'font-size': '20',
                                                            'font-family': font_family,
                                                            'stroke':"black",
                                                            'stroke-width':"1"  })

            n += 1

        # now place arrow at the measurement point
        measurement = Decimal(self.get_field_value("measurement"))
        self._minvalue = maxscale[0]
        self._maxvalue = maxscale[-1]
        centre_string = " " + str(self._cx) + " " + str(self._cy) + ")"
        if measurement >= self._maxvalue:
            self[2].update_attribs({"transform" : "rotate(" + str(self._scale_angle/2) + centre_string})
            return
        if measurement <= self._minvalue:
            self[2].update_attribs({"transform" : "rotate(-" + str(self._scale_angle/2) + centre_string})
            return

        measurement_angle = (measurement - self._minvalue)*self._scale_angle/(self._maxvalue-self._minvalue) - self._scale_angle/2

        rotate_string = "rotate(" + str(measurement_angle) + centre_string
        self[2].update_attribs({"transform" : rotate_string})


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sends scaling factor for mapping measurement to scale"""
        return self._make_fieldvalues(maxvalue=str(self._maxvalue), minvalue=str(self._minvalue))


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
  <path /> <!-- the white backing arc of the scale -->
  <path /> <!-- the scale curved line -->
  <polygon /> <!-- the arrow, with rotation linked to the measurement -->
  <circle /> <!-- the hub of the arrow -->
  <!-- lines and text giving the scale values -->
</g>"""



class Angle(Widget):
    """An svg compass-like angle meter, fitting in a 500x500 space
       Being a circle with scale or either 24, 100 or 360 and arrow
    """

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'transform':FieldArg("text", "", jsonset=True),
                        'font_family':FieldArg("text", "arial"),
                        'measurement':FieldArg("text", "0", jsonset=True),
                        'scale_units':FieldArg("text", "360"),
                       }

    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds the compass, held in a 500 high x 500 wide space"""
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args) 
        # insert a circle at arrow hub, of radius 240
        self[0] = tag.ClosedPart(tag_name='circle', attribs={
                                                       "cx": "250",
                                                       "cy": "250",
                                                       "r": "240",
                                                       "fill":"white",
                                                       "stroke":"black",
                                                       "stroke-width":"4" })

        self[1] = tag.ClosedPart(tag_name='circle', attribs={
                                                       "cx": "250",
                                                       "cy": "250",
                                                       "r": "200",
                                                       "fill":"white",
                                                       "stroke":"black",
                                                       "stroke-width":"2" })

        self[2] = tag.ClosedPart(tag_name='circle', attribs={
                                                       "cx": "250",
                                                       "cy": "250",
                                                       "r": "30",
                                                       "fill":"black",
                                                       "stroke":"black",
                                                       "stroke-width":"2" })

        self[3] = tag.ClosedPart(tag_name='polygon', attribs={
                                                           "fill":"black",
                                                           "stroke":"black",
                                                           "stroke-width":"2",
                                                           "points": "250,65 280,120 260,120 260,400, 240,400 240,120, 220,120" })


    def _build(self, page, ident_list, environ, call_data, lang):
        "create compass"                  
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})
        font_family = self.get_field_value("font_family")
        if not font_family:
            font_family = "arial"
        self._scale_units = self.get_field_value("scale_units")
        measurement = Decimal(self.get_field_value("measurement"))
        rotate_string = ''
        if self._scale_units == '24':
            self._draw_24(font_family)
            # now place arrow at the measurement point
            if (measurement > 0) and (measurement < 24):
                rotate_string = "rotate(" + str(measurement*15) + " 250 250)"
        elif self._scale_units == '100':
            self._draw_100(font_family)
            # now place arrow at the measurement point
            if (measurement > 0) and (measurement < 100):
                rotate_string = "rotate(" + str(measurement*Decimal("3.6")) + " 250 250)"
        else:
            self._scale_units = '360'
            self._draw_360(font_family)
            # now place arrow at the measurement point
            if (measurement > 0) and (measurement < 360):
                rotate_string = "rotate(" + str(measurement) + " 250 250)"
        if rotate_string:
            self[3].update_attribs({"transform" : rotate_string})


    def _draw_24(self, font_family):
        "Draw a 24 hr scale"
        for angle in range(0,24):
            if angle:
                rotate_string = "rotate(" + str(angle*15) + " 250 250)"
                # two digit characters moved to the left a bit more than single digits
                if angle < 10:
                    x = '245'
                else:
                    x = '239'
                self.append( tag.Part(tag_name='text', text=str(angle), attribs={
                                                                'x':x,
                                                                'y': '30',
                                                                'font-size': '20',
                                                                'font-family': font_family,
                                                                'stroke':"black",
                                                                'stroke-width':"1",
                                                                'transform' : rotate_string  })  )
                # major lines
                self.append( tag.ClosedPart(tag_name='line', attribs={
                                                            'x1':'250',
                                                            'y1':'40',
                                                            'x2':'250',
                                                            'y2':'55',
                                                            'stroke':"black",
                                                            'stroke-width':"2",
                                                            'transform' : rotate_string  })  )
                for n in range(1,4):
                    # minor lines
                    minor_rotate = "rotate(" + str(angle*15+n*3.75) + " 250 250)"
                    self.append( tag.ClosedPart(tag_name='line', attribs={
                                                                'x1':'250',
                                                                'y1':'40',
                                                                'x2':'250',
                                                                'y2':'50',
                                                                'stroke':"black",
                                                                'stroke-width':"1",
                                                                'transform' : minor_rotate  })  )
            else:
                self.append( tag.Part(tag_name='text', text="24", attribs={
                                                                'x':'239',
                                                                'y': '30',
                                                                'font-size': '20',
                                                                'font-family': font_family,
                                                                'stroke':"black",
                                                                'stroke-width':"1"  })  )
                self.append( tag.ClosedPart(tag_name='line', attribs={
                                                            'x1':'250',
                                                            'y1':'40',
                                                            'x2':'250',
                                                            'y2':'55',
                                                            'stroke':"black",
                                                            'stroke-width':"2"  })  )
                for n in range(1,4):
                    # minor lines
                    minor_rotate = "rotate(" + str(n*3.75) + " 250 250)"
                    self.append( tag.ClosedPart(tag_name='line', attribs={
                                                                'x1':'250',
                                                                'y1':'40',
                                                                'x2':'250',
                                                                'y2':'50',
                                                                'stroke':"black",
                                                                'stroke-width':"1",
                                                                'transform' : minor_rotate  })  )


    def _draw_100(self, font_family):
        "Draw a 100 hr scale"
        for angle in range(0,100,10):
            if angle:
                rotate_string = "rotate(" + str(angle*3.6) + " 250 250)"
                # two digit characters moved to the left a bit more than single digits
                if angle < 10:
                    x = '245'
                else:
                    x = '239'
                self.append( tag.Part(tag_name='text', text=str(angle), attribs={
                                                                'x':x,
                                                                'y': '30',
                                                                'font-size': '20',
                                                                'font-family': font_family,
                                                                'stroke':"black",
                                                                'stroke-width':"1",
                                                                'transform' : rotate_string  })  )
                # major lines
                self.append( tag.ClosedPart(tag_name='line', attribs={
                                                            'x1':'250',
                                                            'y1':'40',
                                                            'x2':'250',
                                                            'y2':'55',
                                                            'stroke':"black",
                                                            'stroke-width':"2",
                                                            'transform' : rotate_string  })  )
                for n in range(2,10,2):
                    # minor lines
                    minor_rotate = "rotate(" + str((angle+n)*3.6) + " 250 250)"
                    self.append( tag.ClosedPart(tag_name='line', attribs={
                                                                'x1':'250',
                                                                'y1':'40',
                                                                'x2':'250',
                                                                'y2':'50',
                                                                'stroke':"black",
                                                                'stroke-width':"1",
                                                                'transform' : minor_rotate  })  )
                for n in range(1,11,2):
                    # mini lines
                    mini_rotate = "rotate(" + str((angle+n)*3.6) + " 250 250)"
                    self.append( tag.ClosedPart(tag_name='line', attribs={
                                                                'x1':'250',
                                                                'y1':'45',
                                                                'x2':'250',
                                                                'y2':'50',
                                                                'stroke':"black",
                                                                'stroke-width':"1",
                                                                'transform' : mini_rotate  })  )
            else:
                self.append( tag.Part(tag_name='text', text="0", attribs={
                                                                'x':'245',
                                                                'y': '30',
                                                                'font-size': '20',
                                                                'font-family': font_family,
                                                                'stroke':"black",
                                                                'stroke-width':"1"  })  )
                self.append( tag.ClosedPart(tag_name='line', attribs={
                                                            'x1':'250',
                                                            'y1':'40',
                                                            'x2':'250',
                                                            'y2':'55',
                                                            'stroke':"black",
                                                            'stroke-width':"2"  })  )
                for n in range(2,10,2):
                    # minor lines
                    minor_rotate = "rotate(" + str(n*3.6) + " 250 250)"
                    self.append( tag.ClosedPart(tag_name='line', attribs={
                                                                'x1':'250',
                                                                'y1':'40',
                                                                'x2':'250',
                                                                'y2':'50',
                                                                'stroke':"black",
                                                                'stroke-width':"1",
                                                                'transform' : minor_rotate  })  )
                for n in range(1,11,2):
                    # mini lines
                    mini_rotate = "rotate(" + str(n*3.6) + " 250 250)"
                    self.append( tag.ClosedPart(tag_name='line', attribs={
                                                                'x1':'250',
                                                                'y1':'45',
                                                                'x2':'250',
                                                                'y2':'50',
                                                                'stroke':"black",
                                                                'stroke-width':"1",
                                                                'transform' : mini_rotate  })  )

    def _draw_360(self, font_family):
        "Draw a 360 hr scale"
        for angle in range(0,360,15):
            if angle:
                rotate_string = "rotate(" + str(angle) + " 250 250)"
                # two digit characters moved to the left a bit more than single digits
                if angle < 10:
                    x = '245'
                if angle < 100:
                    x = '239'
                else:
                    x = '233'
                self.append( tag.Part(tag_name='text', text=str(angle), attribs={
                                                                'x':x,
                                                                'y': '30',
                                                                'font-size': '20',
                                                                'font-family': font_family,
                                                                'stroke':"black",
                                                                'stroke-width':"1",
                                                                'transform' : rotate_string  })  )
                # major lines
                self.append( tag.ClosedPart(tag_name='line', attribs={
                                                            'x1':'250',
                                                            'y1':'40',
                                                            'x2':'250',
                                                            'y2':'55',
                                                            'stroke':"black",
                                                            'stroke-width':"2",
                                                            'transform' : rotate_string  })  )
                for n in range(5,15,5):
                    # minor lines
                    minor_rotate = "rotate(" + str(angle+n) + " 250 250)"
                    self.append( tag.ClosedPart(tag_name='line', attribs={
                                                                'x1':'250',
                                                                'y1':'40',
                                                                'x2':'250',
                                                                'y2':'50',
                                                                'stroke':"black",
                                                                'stroke-width':"1",
                                                                'transform' : minor_rotate  })  )
            else:
                self.append( tag.Part(tag_name='text', text="0", attribs={
                                                                'x':'245',
                                                                'y': '30',
                                                                'font-size': '20',
                                                                'font-family': font_family,
                                                                'stroke':"black",
                                                                'stroke-width':"1"  })  )
                self.append( tag.ClosedPart(tag_name='line', attribs={
                                                            'x1':'250',
                                                            'y1':'40',
                                                            'x2':'250',
                                                            'y2':'55',
                                                            'stroke':"black",
                                                            'stroke-width':"2"  })  )
                for n in range(5,15,5):
                    # minor lines
                    minor_rotate = "rotate(" + str(n) + " 250 250)"
                    self.append( tag.ClosedPart(tag_name='line', attribs={
                                                                'x1':'250',
                                                                'y1':'40',
                                                                'x2':'250',
                                                                'y2':'50',
                                                                'stroke':"black",
                                                                'stroke-width':"1",
                                                                'transform' : minor_rotate  })  )

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sends scaling factor for mapping measurement to scale"""
        return self._make_fieldvalues(scale_units=self._scale_units)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
  <circle /> <!-- the compass circle -->
  <!-- with further scale and arrow contents -->
</g>"""









"""A Respond object instance is placed in a Respond page,
   The instance is callable, and the respondpage calls it
   to provide the page action"""

import copy, sys, traceback

from .. import skiboot, tag
from ..excepts import ValidateError, ServerError, FailPage, ErrorMessage, PageError, GoTo


class Respond(object):
    """This respond object merely passes form data to the target page, its
main purpose is to act as a parent class for all other respond objects.
"""

    # This indicates if a list of allowed caller idents is required
    allowed_callers_required = False

    # This indicates if a target page ident is required
    target_ident_required = False

    # This indicates if a widgfield is required
    widgfield_required = False

    # This indicates if an alternate page ident is required
    alternate_ident_required = False

    # If True, then a submit list and a fail ident are required
    submit_required = False
    
   # This indicates if the submit option is available
   # if the submit option is chosen, then a submit list and fail ident can also be chosen
    submit_option_available = False
    
    # This indicates if the validate option is available
    validate_option_available = False

    # Options for the fields argument
    field_options = {'fields': True,                  # If False, no fields are expected
                     'widgfields':True,               # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,        # if fields is True, then False here means fields must be supplied
                     'field_values': True,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # False if any number of fields, True if only a single field is required

    def __init__(self, widgfield='', alternate_ident='', target_ident='', fail_ident='', validate_fail_ident='', validate_option=False, submit_option=False, allowed_callers=[], submit_list=[]):
        if self.widgfield_required:
            self.widgfield = skiboot.make_widgfield(widgfield)._replace(i='')
        else:
            self.widgfield = ''
        if self.alternate_ident_required:
            self.alternate_ident = skiboot.make_ident_or_label_or_url(alternate_ident)
        else:
            self.alternate_ident = ''
        if self.target_ident_required:
            self.target_ident = skiboot.make_ident_or_label_or_url(target_ident)
        else:
            self.target_ident = None
        if self.allowed_callers_required:
            self.allowed_callers = [ skiboot.make_ident_or_label(aci) for aci in allowed_callers ]
            if (None in self.allowed_callers) or ('' in self.allowed_callers):
                raise ValidateError("invalid allowed caller ident")
        else:
            self.allowed_callers = []
        if self.submit_option_available:
            self.submit_option = submit_option
        else:
            self.submit_option = False
        if self.submit_required or self.submit_option:
            self.submit_list = submit_list
            self.fail_ident = skiboot.make_ident_or_label_or_url(fail_ident)
        else:
            self.submit_list = []
            self.fail_ident = None
        if self.validate_option_available:
            self.validate_option = validate_option
        else:
            self.validate_option = False
        if self.validate_option:
            self.validate_fail_ident = skiboot.make_ident_or_label_or_url(validate_fail_ident)
        else:
            self.validate_fail_ident = None
        # self.fields is a dictionary of field:value
        self.fields = {}
        # the actual values are inserted when the responder is created
        # by a call to self.set_fields()


    def ident_for_user(self, ident):
        "Returns a tuple if item is ident, or just ident otherwise"
        if isinstance(ident, skiboot.Ident):
            return ident.to_tuple()
        return ident

    @classmethod
    def module_name(cls):
        return cls.__module__.split('.')[-1]


    def set_fields(self, fields):
        """Argument fields is a dictionary of fields and values passed to this
           responder by the user and subject to the field_options.
            This function sets the attribute self.fields to these fields, converting
           the keys and values to skiboot.WidgField objects if the widgfields option is True"""

        if fields:
            # fields are present
            if not self.field_options['fields']:
                raise ServerError(message="No fields are expected")
            if self.field_options['single_field']:
                if len(fields) != 1:
                    raise ServerError(message="Only one field should be given")
        else:
            # no fields present
            if self.field_options['fields'] and (not self.field_options['fields_optional']):
                raise ServerError(message="fields are expected")
            self.fields = {}
            return
        oldfields = self.fields.copy()
        self.fields.clear()
        try:
            for field,value in fields.items():
                self._set_field_value(field, value)
        except ServerError as e:
            self.fields = oldfields
            raise

    def set_field(self, field, value):
        """Sets a single field and value into self.fields"""
        if not field:
            raise ServerError(message="Invalid empty field")
        if not self.field_options['fields']:
            raise ServerError(message="No fields are expected")
        if self.field_options['single_field']:
            oldfields = self.fields.copy()
            self.fields = {}
            try:
                _set_field_value(self, field, value)
            except ServerError as e:
                self.fields = oldfields
                raise
        self._set_field_value(field, value)


    def _set_field_value(self, field, value):
        "Used by above two methods to set the field and value"
        if (not self.field_options['field_values']) and value:
            raise ServerError(message="field values are not expected")
        if not self.field_options['field_values']:
            # ensure False, None 0 are saved as ''
            value = ''
        if self.field_options['field_values'] and (not self.field_options['empty_values_allowed']) and (value==''):
            raise ServerError(message="No field can have an empty value")
        if self.field_options['widgfields']:
            # ensure field is a skiboot.WidgField object
            fld = skiboot.make_widgfield(field)._replace(i='')
        else:
            fld = field
        if self.field_options['widgfield_values']:
            # ensure value is a skiboot.WidgField object
            val = skiboot.make_widgfield(value)._replace(i='')
        else:
            val = value
        self.fields[fld] = val


    @property
    def responder_fields(self):
        """Returns a dictionary of fields to values - if items are widgfields, converts them to comma separated strings"""
        if not self.fields:
            return {}
        string_dict = {}
        for field, value in self.fields.items():
            fld = field.to_str_tuple() if isinstance(field, skiboot.WidgField) else field
            val = value.to_str_tuple() if isinstance(value, skiboot.WidgField) else value
            string_dict[fld] = val
        return string_dict


    def original_fields(self):
        """Returns a dictionary of fields to values - with strings rather than widgfields"""
        if not self.fields:
            return {}
        string_dict = {}
        for field, value in self.fields.items():
            fld = str(field) if isinstance(field, skiboot.WidgField) else field
            val = str(value) if isinstance(value, skiboot.WidgField) else value
            string_dict[fld] = val
        return string_dict


    def original_args(self):
        """Returns a dictionary of arguments - with strings rather than
           widgfields, however idents remain skiboot Ident objects and boolean validate_option remains boolean"""
        string_dict = {}
        if self.widgfield_required:
            string_dict['widgfield'] = str(self.widgfield)
        if self.alternate_ident_required:
            string_dict['alternate_ident'] = self.alternate_ident
        if self.target_ident_required:
            string_dict['target_ident'] = self.target_ident
        if self.submit_required or self.submit_option_available:
            string_dict['fail_ident'] = self.fail_ident
        if self.allowed_callers_required:
            string_dict['allowed_callers'] = self.allowed_callers
        if self.submit_list:
            string_dict['submit_list'] = self.submit_list
        if self.submit_option_available:
            string_dict['submit_option'] = self.submit_option
        if self.validate_option_available:
            string_dict['validate_option'] = self.validate_option
            string_dict['validate_fail_ident'] = self.validate_fail_ident
        return string_dict


    def _submit_data(self, ident_list, skicall):
        "Calls the appropriate user submit_data function"
        # the call could have been passed to another project
        # so update skicall with this project
        proj_ident = ident_list[-1].proj
        this_project = skiboot.getproject(proj_ident)
        skicall.project = proj_ident
        skicall.proj_data = this_project.proj_data
        skicall.rootproject = this_project.rootproject

        # add project brief to submit_dict
        skicall.submit_dict['project_brief'] = this_project.brief
        tuple_ident_list = [ ident.to_tuple() for ident in ident_list ]
        try:
            skicall.ident_list = tuple_ident_list
            skicall.submit_list = self.submit_list.copy()
            result = this_project.submit_data(skicall)
        except (GoTo, FailPage, ServerError, ValidateError) as e:
            raise e
        except Exception:
            message = 'Error in submit_data called by responder ' + str(tuple_ident_list[-1]) + '\n'
            if skiboot.get_debug():
                exc_type, exc_value, exc_traceback = sys.exc_info()
                str_list = traceback.format_exception(exc_type, exc_value, exc_traceback)
                for item in str_list:
                    message += item
            raise ServerError(message)
        return result



    def _check_allowed_callers(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        """Method to check allowed callers, raises a ValidateError if caller not in list of allowed callers
           Only useful for responders that have 'allowed_callers_required'"""
        if not self.allowed_callers_required:
            raise ValidateError(message="Call to _check_allowed_callers in responder, but no allowed_callers_required")
        if not self.allowed_callers: 
            return
        allowed_idents = []
        # convert idents or labels to idents
        for item in self.allowed_callers:
            ident = skiboot.find_ident(item, proj_ident=proj_ident)
            if ident:
                allowed_idents.append(ident)
            else:
                raise ValidateError(message="ident not recognised in responder allowed_callers list")
        if not allowed_idents:
            raise ValidateError(message="No recognised idents given in responder allowed_callers list")
        if len(ident_list) > 1:
            # get previous responder which called this one
            if ident_list[-2] not in allowed_idents:
                raise ValidateError(message="Previouse responder not in this responders allowed_callers list")
        else:
            # not a responder, so check caller_ident
            if caller_page is None:
                raise ValidateError(message="Caller page not known")
            if caller_page.ident not in allowed_idents:
                raise ValidateError(message="Caller page ident not in responder allowed_callers list")


    def _validate_fields(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        "Validates the fields specified in self.fields, returns validated form data, but does not change original form_data"
        # validation tests are stored in caller page
        if caller_page is None:
            raise ValidateError()
        if not self.field_options['widgfields']:
            raise ValidateError()
                 
       # create a copy of form_data which will hold validated data
        validated_form_data = form_data.copy()
        
        # for each field and value, validate the value, placing data into validated_form_data
        # If any value fails, then place any errors into e_list
        # e_list is a list of ErrorMessage exceptions with message to be displayed, and where to display them
        e_list = []
        # error_dict is a dictionary of errored widgfields: original value
        error_dict = {}

        for field in self.fields:
            # validate each field
            if field not in form_data:
                validated_form_data[field] = ''
            value = validated_form_data[field]
            validated_form_data[field], errors = caller_page.validate(field, value, skicall.environ, skicall.lang, validated_form_data, skicall.call_data, skicall.page_data)
            if errors:
                e_list.extend(errors)
                if field.s:
                    error_dict[field.s, field.w, field.f] = value
                else:
                    error_dict[field.w, field.f] = value

        # validated_form_data now holds new values
        # e_list holds errors occurred
        if e_list:
            # on validation errors, call the validate_fail_ident page, get final target template, and show the errors
            page = self.get_page_from_ident(self.validate_fail_ident, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata, error_dict)
            raise PageError(page, e_list)
        # so all ok, no error, but some values may be substituted, so return the validated form data
        return validated_form_data

    def get_page_from_ident(self, ident, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata, error_dict=None):
        """Calls the next responder or template page given by ident (which can be an ident, label or url)
           - and if a responder, calls its respond object to finally return a final target page."""
        if isinstance(ident, str) and ('/' in ident):
            # this is a URL
            return ident
        # ident is either a string ident or label
        thisident = skiboot.find_ident_or_url(ident, proj_ident)
        # thisident is either None, an Ident object or a url
        if thisident is None:
            raise ValidateError()
        if isinstance(thisident, str):
            # this is a URL
            return thisident
        # so thisident is an Ident object
        page = thisident.item()
        if not page:
            raise ValidateError()
        # could be a folder
        if page.page_type == 'Folder':
            page = page.default_page
            if not page:
                raise ValidateError()
        if hasattr(page, 'page_type') and page.page_type == 'RespondPage':
            # must call the pages respond object
            page = page.call_responder(skicall, form_data, caller_page, ident_list, rawformdata, error_dict)
        return page

 
    def get_target_page(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        return self.get_page_from_ident(self.target_ident, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)


    def get_fail_page(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        return self.get_page_from_ident(self.fail_ident, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)


    def get_alternate_page(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        "Gets the alternate page, if the alternate page has an ident, if it is an external url, get the redirector page with this url"
        return self.get_page_from_ident(self.alternate_ident, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)


    def raise_error_page(self, e_list, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        """Gets the page with fail_ident, sets error_messages and raises a PageError holding the page
           e_list is a list of ErrorMessage instances"""
        try:
            page = self.get_page_from_ident(self.fail_ident, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        except GoTo as e:
            e.e_list = e_list
            raise e  
        raise PageError(page, e_list)


    def __call__(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata, error_dict=None):
        "gets the project ident, and page messages and calls self._respond"

        if self.target_ident_required:
            skicall.submit_dict['target_ident'] = self.ident_for_user(self.target_ident)
        if self.submit_required or self.submit_option_available:
            skicall.submit_dict['fail_ident'] = self.ident_for_user(self.fail_ident)
        if self.alternate_ident_required:
            skicall.submit_dict['alternate_ident'] = self.ident_for_user(self.alternate_ident)
        # call self._respond
        try:
            page = self._respond(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        except GoTo as e:
            e.proj_ident=proj_ident
            raise e
        except ValidateError:
            raise
        except ServerError:
            raise
        except PageError:
            raise
        # Any other error, raise ServerError
        except Exception:
            message = "Uncaught exception in user code responder %s,%s" % ident_list[-1]
            if skiboot.get_debug():
                message += "\n"
                exc_type, exc_value, exc_traceback = sys.exc_info()
                str_list = traceback.format_exception(exc_type, exc_value, exc_traceback)
                for item in str_list:
                    message += item
            raise ServerError(message)
        return page


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        """Should be overridden
        this method then returns the target page - or the ultimate page
        if the target is itself another Respond page
        form_data is a dictionary"""
        # return the target page
        try:
            return self.get_target_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        except FailPage as e:
            # raises a PageError exception
            self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)



# import all responders here

from .checkers import AllowedFields, PrettyFormData, StoreData, StoreDataKeyed, AllowStoreKeyed, AllowStore, Accept, AllowedAccept, PageData
from .submitters import SubmitData, SubmitCSS, GetDictionaryDefaults, FieldStoreSubmit, ColourSubstitute, SetCookies, SubmitJSON, SubmitPlainText, SubmitIterator, MediaQuery
from .navigators import CaseSwitch, EmptyGoto, EmptyCallDataGoto, DelCallDataItem, NoOperation






"Functions implementing validator editing"

from skipole import ServerError, FailPage, ValidateError, GoTo

from skipole import skilift
from skipole.skilift import editwidget, editvalidator


def _field_name(widget, field_argument):
    "Returns a field name"
    if "set_names" not in widget:
        return field_argument
    name_dict = widget["set_names"]
    if field_argument in name_dict:
        return name_dict[field_argument]
    return field_argument


def retrieve_editvalidator(skicall):
    "Fills in the edit a validator page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    if section_name:
        page_data[("validator_displaywidget_textblock","replace_strings")] = ['If the widget is in this section, the name should be of the form %s,widget_name.' % (section_name,)]
    else:
        page_data[("validator_displaywidget_textblock","replace_strings")] = ['If the widget is in a section, the name should be of the form section_alias,widget_name.']

    # get validator
    try:
        validx = int(call_data['validx'])
    except Exception:
        raise FailPage("Unknown validator to edit")

    # get validator and widget info
    try:
        if section_name:
            widgetdescription = editwidget.section_widget_description(project, section_name, call_data['schange'], widget_name)
            widget =  editwidget.section_widget(project, section_name, call_data['schange'], widget_name)
            vinfo = editvalidator.section_field_validator_info(project, section_name, call_data['schange'], widget_name, field_arg, validx)
        else:
            widgetdescription = editwidget.page_widget_description(project, pagenumber, call_data['pchange'], widget_name)
            widget = editwidget.page_widget(project, pagenumber, call_data['pchange'], widget_name)
            vinfo = editvalidator.page_field_validator_info(project, pagenumber, call_data['pchange'], widget_name, field_arg, validx)
    except ServerError as e:
        raise FailPage(e.message)

    field_name = _field_name(widget, field_arg)

    call_data['extend_nav_buttons'].append(["back_to_field_edit", "Back to field", True, ''])

    page_data[("adminhead","page_head","large_text")] = "Edit : %s on field %s" % (vinfo.validator, field_name)

    page_data[('widget_type','para_text')] = "Widget type : %s.%s" % (widgetdescription.modulename, widgetdescription.classname)
    page_data[('widget_name','para_text')] = "Widget name : %s" % (widget_name,)
    page_data[('field_type','para_text')] = "Field type : %s" % (field_arg,)
    page_data[('field_name','para_text')] = "Field name : %s" % (field_name,)
    page_data[('validator_type','para_text')] = "Validator type : %s.%s" % (vinfo.module_name,vinfo.validator)

    page_data[('validator_textblock','textblock_ref')] = ".".join(("validators",vinfo.module_name,vinfo.validator))

    page_data[('e_message','input_text')] = vinfo.message
    page_data[('e_message_ref','input_text')] = vinfo.message_ref

    page_data[('displaywidget','input_text')] = vinfo.displaywidget

    # list of allowed values
    contents = []
    allowed_vals = vinfo.allowed_values
    for idx, val in enumerate(allowed_vals):
        row = [val, str(idx)]
        contents.append(row)
    if contents:
        page_data[('allowed_values','contents')] = contents
        page_data[('allowed_values','show')] = True
    else:
        page_data[('allowed_values','show')] = False

    # Validator arguments
    arg_contents = []
    val_args = vinfo.val_args
    for name, value in val_args.items():
        row = [name, str(value), name]
        arg_contents.append(row)
    if arg_contents:
        arg_contents.sort(key=lambda x: x[0])
        page_data[('validator_args','contents')] = arg_contents
        page_data[('validator_args','show')] = True
        page_data[('description5','show')] = True
    else:
        page_data[('validator_args','show')] = False
        page_data[('description5','show')] = False


def set_e_message(skicall):
    "Sets a validator error message"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    if 'e_message' not in call_data:
        raise FailPage("Error message not given")

    if call_data['e_message']:
        e_message = call_data['e_message']
    else:
        e_message = ''

    # get validator index
    try:
        validx = int(call_data['validx'])
    except Exception:
        raise FailPage("Invalid validator")

    # set message
    try:
        if section_name:
            call_data['schange'] = editvalidator.set_section_field_validator_error_message(project, section_name, call_data['schange'], widget_name, field_arg, validx, e_message)
        else:
            call_data['pchange'] = editvalidator.set_page_field_validator_error_message(project, pagenumber, call_data['pchange'], widget_name, field_arg, validx, e_message)
    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = "Validator error message changed"


def set_e_message_ref(skicall):
    "Sets a validator error message reference"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    if 'e_message_ref' not in call_data:
        raise FailPage("Error message reference not given")

    if call_data['e_message_ref']:
        e_message_ref = call_data['e_message_ref']
    else:
        e_message_ref = ''

    # get validator index
    try:
        validx = int(call_data['validx'])
    except Exception:
        raise FailPage("Invalid validator")

    # set message reference
    try:
        if section_name:
            call_data['schange'] = editvalidator.set_section_field_validator_error_message_reference(project, section_name, call_data['schange'], widget_name, field_arg, validx, e_message_ref)
        else:
            call_data['pchange'] = editvalidator.set_page_field_validator_error_message_reference(project, pagenumber, call_data['pchange'], widget_name, field_arg, validx, e_message_ref)
    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = "Validator error message reference changed"


def set_displaywidget(skicall):
    "Sets a validator displaywidget"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    if 'displaywidget' not in call_data:
        raise FailPage("Display widget not given")

    if call_data['displaywidget']:
        displaywidget = call_data['displaywidget']
    else:
        displaywidget = ''

    # get validator index
    try:
        validx = int(call_data['validx'])
    except Exception:
        raise FailPage("Invalid validator")

    # set displaywidge
    try:
        if section_name:
            call_data['schange'] = editvalidator.set_section_field_validator_displaywidget(project, section_name, call_data['schange'], widget_name, field_arg, validx, displaywidget)
        else:
            call_data['pchange'] = editvalidator.set_page_field_validator_displaywidget(project, pagenumber, call_data['pchange'], widget_name, field_arg, validx, displaywidget)
    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = "Validator display widget changed"


def set_allowed_value(skicall):
    "Adds a validator allowed value"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    # get validator index
    try:
        validx = int(call_data['validx'])
    except Exception:
        raise FailPage("Invalid validator")

    if 'add_allowed' not in call_data:
        raise FailPage("Allowed value to be added not given")

    if call_data['add_allowed']:
        allowed_value = call_data['add_allowed']
    else:
        raise FailPage("A none-empty string is required")

    # add allowed_value
    try:
        if section_name:
            call_data['schange'] = editvalidator.add_section_field_validator_allowed_value(project, section_name, call_data['schange'], widget_name, field_arg, validx, allowed_value)
        else:
            call_data['pchange'] = editvalidator.add_page_field_validator_allowed_value(project, pagenumber, call_data['pchange'], widget_name, field_arg, validx, allowed_value)
    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = "Validator allowed values changed"


def remove_allowed_value(skicall):
    "Removes a validator allowed value"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    # get validator index
    try:
        validx = int(call_data['validx'])
    except Exception:
        raise FailPage("Invalid validator")

    if 'remove_allowed' not in call_data:
        raise FailPage("Allowed value to be removed not given")

    try:
        idx = int(call_data['remove_allowed'])
    except Exception:
        raise FailPage("Invalid allowed value")

    # remove allowed_value
    try:
        if section_name:
            call_data['schange'] = editvalidator.remove_section_field_validator_allowed_value(project, section_name, call_data['schange'], widget_name, field_arg, validx, idx)
        else:
            call_data['pchange'] = editvalidator.remove_page_field_validator_allowed_value(project, pagenumber, call_data['pchange'], widget_name, field_arg, validx, idx)
    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = "Validator allowed values changed"


def retrieve_arg(skicall):
    "Fills in the edit a validator argument page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    # get validator
    try:
        validx = int(call_data['validx'])
    except Exception:
        raise FailPage("Unknown validator to edit")

    if 'arg_name' not in call_data:
        raise FailPage("Validator argument not identified")
    arg_name = call_data['arg_name']

    # get validator and widget info
    try:
        if section_name:
            widgetdescription = editwidget.section_widget_description(project, section_name, call_data['schange'], widget_name)
            widget =  editwidget.section_widget(project, section_name, call_data['schange'], widget_name)
            vinfo = editvalidator.section_field_validator_info(project, section_name, call_data['schange'], widget_name, field_arg, validx)
            page_data[("validator_displaywidget_textblock","replace_strings")] = ['If the widget is in this section, the name should be of the form %s,widget_name.' % (section_name,)]
        else:
            widgetdescription = editwidget.page_widget_description(project, pagenumber, call_data['pchange'], widget_name)
            widget = editwidget.page_widget(project, pagenumber, call_data['pchange'], widget_name)
            vinfo = editvalidator.page_field_validator_info(project, pagenumber, call_data['pchange'], widget_name, field_arg, validx)
            page_data[("validator_displaywidget_textblock","replace_strings")] = ['If the widget is in a section, the name should be of the form section_alias,widget_name.']
    except ServerError as e:
        raise FailPage(e.message)

    if arg_name not in vinfo.val_args:
        raise FailPage("Validator argument not identified")

    field_name = _field_name(widget, field_arg)

    call_data['extend_nav_buttons'].extend([["back_to_field_edit", "Back to field", True, ''],["back_to_validator", "Back to validator", True, '']])

    # navigator text
    page_data[("adminhead","page_head","large_text")] = arg_name

    page_data[('widget_type','para_text')] = "Widget type : %s.%s" % (widgetdescription.modulename, widgetdescription.classname)
    page_data[('widget_name','para_text')] = "Widget name : %s" % (widget_name,)
    page_data[('field_type','para_text')] = "Field type : %s" % (field_arg,)
    page_data[('field_name','para_text')] = "Field name : %s" % (field_name,)
    page_data[('validator_type','para_text')] = "Validator type : %s.%s" % (vinfo.module_name,vinfo.validator)

    page_data[('validator_arg_textblock','textblock_ref')] = ".".join(("validators",vinfo.module_name,vinfo.validator,arg_name))

    page_data[('arg_val','input_text')] = vinfo.val_args[arg_name]
    page_data[('arg_val','hidden_field1')] = arg_name


def set_arg_value(skicall):
    "Sets a validator argument value"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    # get validator
    try:
        validx = int(call_data['validx'])
    except Exception:
        raise FailPage("Unknown validator to edit")

    if 'arg_name' not in call_data:
        raise FailPage("argument name not given")
    if 'arg_value' not in call_data:
        raise FailPage("argument value not given")

    arg_name = call_data['arg_name']
    arg_value = call_data['arg_value']

    try:
        if section_name:
            call_data['schange'] = editvalidator.set_section_field_validator_argument(project, section_name, call_data['schange'], widget_name, field_arg, validx, arg_name, arg_value)
        else:
            call_data['pchange'] = editvalidator.set_page_field_validator_argument(project, pagenumber, call_data['pchange'], widget_name, field_arg, validx, arg_name, arg_value)
    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = "Validator argument changed"


def move_up(skicall):
    "Moves a validator up in a field validator list"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    # get validator
    try:
        validx = int(call_data['validx'])
    except Exception:
        raise FailPage("Invalid value to move")

    if validx == 0:
        raise FailPage("Invalid value to move")

    # move validator
    try:
        if section_name:
            call_data['schange'] = editvalidator.swap_section_field_validators(project, section_name, call_data['schange'], widget_name, field_arg, validx, validx-1)
        else:
            call_data['pchange'] = editvalidator.swap_page_field_validators(project, pagenumber, call_data['pchange'], widget_name, field_arg, validx, validx-1)
    except ServerError as e:
        raise FailPage(e.message)
    del call_data['validx']


def move_down(skicall):
    "Moves a validator down in a field validator list"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    # get validator
    try:
        validx = int(call_data['validx'])
    except Exception:
        raise FailPage("Invalid value to move")

    # move validator
    try:
        if section_name:
            call_data['schange'] = editvalidator.swap_section_field_validators(project, section_name, call_data['schange'], widget_name, field_arg, validx, validx+1)
        else:
            call_data['pchange'] = editvalidator.swap_page_field_validators(project, pagenumber, call_data['pchange'], widget_name, field_arg, validx, validx+1)
    except ServerError as e:
        raise FailPage(e.message)
    del call_data['validx']


def remove_validator(skicall):
    "Removes a validator"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    # get validator
    try:
        validx = int(call_data['validx'])
    except Exception:
        raise FailPage("Invalid value to remove")

    # remove validator
    try:
        if section_name:
            call_data['schange'] = editvalidator.remove_section_field_validator(project, section_name, call_data['schange'], widget_name, field_arg, validx)
        else:
            call_data['pchange'] = editvalidator.remove_page_field_validator(project, pagenumber, call_data['pchange'], widget_name, field_arg, validx)
    except ServerError as e:
        raise FailPage(e.message)

    del call_data['validx']
    call_data['status'] = "Validator removed"


def retrieve_validator_modules(skicall):
    "Creates a list of validator modules"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    # get a WidgetDescription named tuple, and a widget dictionary

    try:
        if section_name:
            widgetdescription = editwidget.section_widget_description(project, section_name, call_data['schange'], widget_name)
            widget =  editwidget.section_widget(project, section_name, call_data['schange'], widget_name)
        else:
            widgetdescription = editwidget.page_widget_description(project, pagenumber, call_data['pchange'], widget_name)
            widget = editwidget.page_widget(project, pagenumber, call_data['pchange'], widget_name)
    except ServerError as e:
        raise FailPage(e.message)

    field_name = _field_name(widget, field_arg)

    page_data[("adminhead","page_head","large_text")] = "Add validator to (\"%s\",\"%s\")" % (widget_name, field_name)

    page_data[('widget_type','para_text')] = "Widget type : %s.%s" % (widgetdescription.modulename, widgetdescription.classname)
    page_data[('widget_name','para_text')] = "Widget name : %s" % (widget_name,)
    page_data[('field_type','para_text')] = "Field type : %s" % (field_arg,)
    page_data[('field_name','para_text')] = "Field name : %s" % (field_name,)

    call_data['extend_nav_buttons'].append(["back_to_field_edit", "Back to field", True, ''])

    # table of validator modules

    # col 0 is the visible text to place in the link,
    # col 1 is the get field of the link
    # col 2 is the get field of the link
    # col 3 is the reference string of a textblock to appear in the column adjacent to the link
    # col 4 is text to appear if the reference cannot be found in the database
    # col 5 normally empty string, if set to text it will replace the textblock

    validator_modules = editvalidator.validator_modules()

    page_data[("modulestable","link_table")] = [ [name, name, '', ".".join(('validators', name, 'module')), 'Description not found', ''] for name in validator_modules ]


def retrieve_validator_list(skicall):
    "Creates a list of validators"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")

    # get a WidgetDescription named tuple, and a widget dictionary

    try:
        if section_name:
            widgetdescription = editwidget.section_widget_description(project, section_name, call_data['schange'], widget_name)
            widget =  editwidget.section_widget(project, section_name, call_data['schange'], widget_name)
        else:
            widgetdescription = editwidget.page_widget_description(project, pagenumber, call_data['pchange'], widget_name)
            widget = editwidget.page_widget(project, pagenumber, call_data['pchange'], widget_name)
    except ServerError as e:
        raise FailPage(e.message)

    field_name = _field_name(widget, field_arg)

    page_data[("adminhead","page_head","large_text")] = "Add validator to (\"%s\",\"%s\")" % (widget_name, field_name)

    page_data[('widget_type','para_text')] = "Widget type : %s.%s" % (widgetdescription.modulename, widgetdescription.classname)
    page_data[('widget_name','para_text')] = "Widget name : %s" % (widget_name,)
    page_data[('field_type','para_text')] = "Field type : %s" % (field_arg,)
    page_data[('field_name','para_text')] = "Field name : %s" % (field_name,)

    call_data['extend_nav_buttons'].extend([["back_to_field_edit", "Back to field", True, ''],["validator_modules", "Modules", True, '']])

    if 'valmodule' in call_data:
        module_name = call_data['valmodule']
    else:
        raise FailPage("Module not identified")

    validator_modules = editvalidator.validator_modules()

    if module_name not in validator_modules:
        raise FailPage("Module not identified")

    page_data[('moduledesc','textblock_ref')] = 'validators.' + module_name + '.module'

    # table of validators

    # col 0 is the visible text to place in the link,
    # col 1 is the get field of the link
    # col 2 is the get field of the link
    # col 3 is the reference string of a textblock to appear in the column adjacent to the link
    # col 4 is text to appear if the reference cannot be found in the database
    # col 5 normally empty string, if set to text it will replace the textblock

    # tuple of validators in the module
    validators = editvalidator.validators_in_module(module_name)
    page_data[("validators","link_table")] = [ [name, name, module_name, ".".join(("validators",module_name,name)), 'Description not found', ''] for name in validators ]


def create_validator(skicall):
    "Creates a validator and adds it to the field"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    section_name = None
    pagenumber = None
    if 'section_name' in call_data:
        section_name = call_data['section_name']
    elif 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message="No section or page given")

    if 'widget_name' in call_data:
        widget_name = call_data['widget_name']
    else:
        raise FailPage(message="Widget not identified")

    if 'field_arg' in call_data:
        field_arg = call_data['field_arg']
    else:
        raise FailPage("Field not identified")
    if 'validator' not in call_data:
        raise FailPage("Validator not given")
    validator_name = call_data['validator']
    if not validator_name:
        raise FailPage("Validator not given")

    if 'valmodule' not in call_data:
        raise FailPage("Validator module not given")

    module_name = call_data['valmodule']

    try:
        if section_name:
            call_data['schange'] =  editvalidator.create_section_field_validator(project, section_name, call_data['schange'], widget_name, field_arg, module_name, validator_name)
            # get new list of validators attached to the field
            val_list = editvalidator.get_section_field_validator_list(project, section_name, call_data['schange'], widget_name, field_arg)
        else:
            call_data['pchange'] =  editvalidator.create_page_field_validator(project, pagenumber, call_data['pchange'], widget_name, field_arg, module_name, validator_name)
            val_list = editvalidator.get_page_field_validator_list(project, pagenumber, call_data['pchange'], widget_name, field_arg)
    except ServerError as e:
        raise FailPage(e.message)
 
    call_data['validx'] = str(len(val_list)-1)
    del call_data['validator']
    del call_data['valmodule']

    call_data['status'] = "Validator added"






import re

# a search for anything none-alphanumeric and not an underscore
_TB = re.compile('[^\w\.]')

from skipole import FailPage, GoTo, skilift
from .. import utils


def retrieve_link_table(skicall):
    "Gets data for the manage textblocks page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # clears any session data
    utils.clear_call_data(call_data)

    project = call_data['editedprojname']

    page_data[("adminhead","page_head","large_text")] = "Manage TextBlocks"

    accesstextblocks = skilift.get_accesstextblocks(project)

    result = accesstextblocks.get_textrefs()
    # result is a list of textrefs

    #       contents: col 0 is the text to place in the first column,
    #                  col 1, 2, 3, 4 are the get field contents of links 1,2,3 and 4
    #                  col 5 - True if the first button and link is to be shown, False if not
    #                  col 6 - True if the second button and link is to be shown, False if not
    #                  col 7 - True if the third button and link is to be shown, False if not
    #                  col 8 - True if the fourth button and link is to be shown, False if not

    page_data[('txtblocks', 'contents')] = [ [ref, ref, ref, '', '', True, True, False, False] for ref in result if '.' not in ref]
    if not page_data[('txtblocks', 'contents')]:
        page_data[('txtblocks', 'show')] = False
    else:
        page_data[('txtblocks', 'contents')].sort(key=lambda r: r[0])

    # set page data for widget tables000.Table1_Button name moretblocks - displaying refs that do have a '.' in them
    # contents: col 0 is the text to place in the first column, col 1 is the get field

    reflinklist = []
    for ref in result:
        if '.' not in ref:
            continue
        ritem = ref.split('.')[0]
        if ritem not in reflinklist:
            reflinklist.append(ritem)


    if not reflinklist:
        page_data[('moretblocks', 'show')] = False
        return
    
    page_data[('moretblocks', 'contents')] = [ [ref + "...", ref] for ref in reflinklist]
    page_data[('moretblocks', 'contents')].sort(key=lambda r: r[0])


def retrieve_more(skicall):
    "Gets data for further textblock tables"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    page_data[("adminhead","page_head","large_text")] = "Further TextBlocks"
    call_data['extend_nav_buttons'] = []

    reference = call_data['ref']
    reflist = reference.split('.')
    refdot = reference + '.'

    # a back button is added if ref has a . in it
    if '.' in reference:
        backref = '.'.join(reflist[:-1])
        call_data['extend_nav_buttons'].append([4006, "Previous", True, backref])

    # set input text
    page_data[("st1","input_text")] = refdot

    # list of textrefs
    accesstextblocks = skilift.get_accesstextblocks(project)
    result = accesstextblocks.get_textrefs()           

    # set result to a list of textrefs that start with refdot
    result = [ ref for ref in result if ref.startswith(refdot) ]

    if not result:
        raise FailPage("No further references found")

   #       contents: col 0 is the text to place in the first column,
    #                  col 1, 2, 3, 4 are the get field contents of links 1,2,3 and 4
    #                  col 5 - True if the first button and link is to be shown, False if not
    #                  col 6 - True if the second button and link is to be shown, False if not
    #                  col 7 - True if the third button and link is to be shown, False if not
    #                  col 8 - True if the fourth button and link is to be shown, False if not


    displaylen = len(reflist) + 1

    newresult = [ ref for ref in result if displaylen == len(ref.split('.')) ]

    if newresult:
        page_data[('txtblocks', 'contents')] = [ [ref, ref,  ref, '', '',  True, True, False, False] for ref in newresult]
        page_data[('txtblocks', 'contents')].sort(key=lambda r: r[0])
    else:
        page_data[('txtblocks', 'show')] = False

    # The more table displays remaining refs in result

    # set page data for widget tables000.Table1_Button name moretblocks - displaying remaining refs
    # contents: col 0 is the text to place in the first column, col 1 is the get_field
 
    reflinklist = []
    for ref in result:
        if ref in newresult:
            continue
        rlist = ref.split('.')
        ritem = '.'.join(rlist[:displaylen])
        if ritem not in reflinklist:
            reflinklist.append(ritem)

    if not reflinklist:
        page_data[('moretblocks', 'show')] = False
        return
    
    page_data[('moretblocks', 'contents')] = [ [ref + "...", ref] for ref in reflinklist]
    page_data[('moretblocks', 'contents')].sort(key=lambda r: r[0])


def retrieve_textblock(skicall):
    "Gets data for the edit textblock page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # clears any session data
    utils.clear_call_data(call_data, keep=['textblock','text','language','status'] )

    project = call_data['editedprojname']
    language = skicall.lang[0].lower()
    default_language = skicall.lang[1]

    page_data[("adminhead","page_head","large_text")] = "Edit TextBlock"
    call_data['extend_nav_buttons'] = []

    if 'textblock' in call_data:
        edited_textblock = call_data['textblock']
    else:
        raise FailPage("TextBlock Reference not found")

    accesstextblocks = skilift.get_accesstextblocks(project)
    result = accesstextblocks.get_textref_languages()
    # result is a dictionary {textref: [languages],...}
    if not edited_textblock in result:
        raise FailPage("The TextBlock reference has not been found")

    if '.' in edited_textblock:
        reflist = edited_textblock.split('.')
        backref = '.'.join(reflist[:-1])
        call_data['extend_nav_buttons'].append([4006, "List TextBlocks", True, backref])
    else:
        call_data['extend_nav_buttons'].append(['manage_textblocks', "List TextBlocks", True, ''])

    page_data['st1:hidden_field1'] = edited_textblock
    page_data['sta1:hidden_field1'] = edited_textblock
    page_data['sb1:hidden_field1'] = edited_textblock
    page_data['sb2:hidden_field1'] = edited_textblock
    page_data['st2:hidden_field1'] = edited_textblock
    page_data[("adminhead","page_head","large_text")] = "Edit TextBlock : " + edited_textblock
    page_data['ts1:para_text'] = _list_to_string(result[edited_textblock])
    # get default language textblock_lang
    langsplit = language.split('-')
    textblock_lang_list = result[edited_textblock]
    if not textblock_lang_list:
        raise FailPage("No language has been found for the TextBlock reference")
    if accesstextblocks.default_language in textblock_lang_list:
        textblock_lang = accesstextblocks.default_language
    elif language in textblock_lang_list:
        textblock_lang = language
    elif langsplit[0] in textblock_lang_list:
        textblock_lang = langsplit[0]
    elif default_language in textblock_lang_list:
        textblock_lang = default_language
    elif 'en' in textblock_lang_list:
        textblock_lang = 'en'
    else:
        textblock_lang = textblock_lang_list[0]
    if not textblock_lang:
        raise FailPage("No language has been found for the TextBlock reference")
    # get submitted textblock language
    if 'language' in call_data:
        textblock_language = call_data['language'].lower()
        textblock_text = accesstextblocks.get_exact_text(edited_textblock, textblock_language)
        if textblock_text is None:
            page_data['sp1:para_text'] = "No text in this language has been found, edit and submit the text below to set this language text\n"
    else:
        textblock_language = textblock_lang
        textblock_text = accesstextblocks.get_exact_text(edited_textblock, textblock_language)
        if textblock_text is None:
            raise FailPage("Failed to find text for the TextBlock reference")

    page_data['st1:input_text'] = textblock_language
    page_data['sta1:hidden_field2'] = textblock_language
    page_data['sb1:hidden_field2'] = textblock_language
    if 'text' in call_data:
        input_text = call_data['text']
    else:
        input_text = accesstextblocks.get_text(edited_textblock, (textblock_language,default_language))
    if input_text is None:
        page_data['sta1:input_text'] = ''
    else:
        page_data['sta1:input_text'] = input_text
    # should delete language be shown
    if (len(textblock_lang_list) < 2) or (textblock_text is None):
        page_data['sb1:show'] = False
    else:
        page_data['sb1:show'] = True


def submit_new_textblock(skicall):
    "A new textblock is to be created"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    if "new_textblock_ref" not in call_data:
        raise ValidateError(message='Invalid call')

    new_textblock_ref = call_data['new_textblock_ref']
    if not new_textblock_ref:
        raise FailPage(message="Invalid reference")

    if _TB.search(new_textblock_ref):
        raise FailPage(message="Invalid reference")

    accesstextblocks = skilift.get_accesstextblocks(project)
    result = accesstextblocks.get_textref_languages()
    # result is a dictionary {textref: [languages],...}
    if not new_textblock_ref in result:
        # A new TextBlock is to be created
        try:
            accesstextblocks.set_text("Insert text here", new_textblock_ref, accesstextblocks.default_language)
        except Exception:
            raise FailPage(message = "Failed to write to the textblocks database")
    # store the reference, as this goes to textblock edit responder which requires this data
    call_data['textblock'] = new_textblock_ref


def submit_text(skicall):
    "Set the text of the textblock"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    if "textblock" not in call_data:
        raise ValidateError(message='Invalid call - no textblock reference')
    if "language" not in call_data:
        raise ValidateError(message='Invalid call - no language set')
    if "text" not in call_data:
        raise ValidateError(message='Invalid call - no text set')
    # set the text and language in this textblock text
    accesstextblocks = skilift.get_accesstextblocks(project)
    accesstextblocks.set_text(call_data['text'], call_data['textblock'], call_data['language'])
    call_data['status'] = "TextBlock updated"


def submit_delete_language(skicall):
    "Delete a language"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    if "textblock" not in call_data:
        raise ValidateError(message='Invalid call - no textblock reference')
    if "delete_language" not in call_data:
        raise ValidateError(message='Invalid call - no language to delete set')
    # request a language deletion
    accesstextblocks = skilift.get_accesstextblocks(project)
    result = accesstextblocks.get_textref_languages()
    # result is a dictionary {textref: [languages],...}
    if call_data['textblock'] not in result:
        raise FailPage("Failed to find the TextBlock reference")
    lang_list = result[call_data['textblock']]
    if call_data['delete_language'] not in lang_list:
        raise FailPage("Failed to find the TextBlock language")
    language = call_data['delete_language'].lower()
    if len(lang_list) < 2:
        raise FailPage("Only one language set, cannot be deleted")
    # delete the block language
    accesstextblocks.del_text(call_data['textblock'], language)
    call_data['status'] = "Language deleted"


def submit_delete_textblock(skicall):
    "Delete a textblock"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    if "delete_textblock" not in call_data:
        raise ValidateError(message='Invalid call - no textblock reference')
    textref = call_data['delete_textblock']
    accesstextblocks = skilift.get_accesstextblocks(project)
    if not accesstextblocks.textref_exists(textref):
        call_data['status'] = "TextBlock not found"
        return
    accesstextblocks.del_textblock(textref)
    call_data['status'] = "TextBlock deleted"
    _find_previous_textblock(skicall, textref)


def _find_previous_textblock(skicall, textref):
    "raises a GoTo if a previous textref exists"
    if '.' not in textref:
        return
    project = skicall.call_data['editedprojname']
    reflist = textref.split('.')
    newref = '.'.join(reflist[:-1])
    # do any newref.* exist ?
    refdot = newref + '.'
    accesstextblocks = skilift.get_accesstextblocks(project)
    # set result to a list of textrefs that start with refdot
    result = [ ref for ref in accesstextblocks.get_textrefs() if ref.startswith(refdot) ]
    if not result:
        return _find_previous_textblock(skicall, newref)
    # some do exist    
    skicall.call_data['ref'] = newref
    # fill in, then go to the template page
    retrieve_more(skicall)
    raise GoTo(24005)



def submit_copy_textblock(skicall):
    "Copy a textblock"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    if "textblock" not in call_data:
        raise ValidateError(message='Invalid call - no textblock reference')
    if "copy_textblock_ref" not in call_data:
        raise ValidateError(message='Invalid call - no new textblock reference')
    if _TB.search(call_data['copy_textblock_ref']):
        raise FailPage(message="Invalid reference")
    accesstextblocks = skilift.get_accesstextblocks(project)
    accesstextblocks.copy(call_data['textblock'], call_data['copy_textblock_ref'])
    call_data['status'] = "TextBlock ref %s copied to %s" % (call_data['textblock'], call_data['copy_textblock_ref'])


def _list_to_string(lang_list):
    string_result = ""
    if not lang_list:
        return ''
    for lang in lang_list:
        string_result += lang
        string_result += ", "
    return string_result[:-2]






"""Contains widgets for radio forms"""


from .. import skiboot, tag, excepts
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict



class RadioButton1(Widget):
    """A radiobutton widget. Without a form or submit button, typically included within a form"""

    error_location = (0,0,0)

    arg_descriptions = {'radio_values':FieldArgList("text"),
                       'radio_text':FieldArgList("text"),
                       'radio_checked':FieldArg("text", '', valdt=True, jsonset=True),
                       'error_class':FieldArg("cssclass", ""),
                       'inputdiv_class':FieldArg("cssclass", ""),
                       'label_class':FieldArg("cssclass", ""),
                       'div_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        radio_values: list of options, one of these values will be returned
        radio_text: List of labels
        radio_checked: value to be checked, this is the field returned
        show: True if this widget is to be visible
        error_class: css class applied to the normally hidden error div
        inputdiv_class: css class applied to each radio div
        label_class: css class applied to each label text on the right of the radio button
        div_class: css class applied to the div after the error paragraph
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # hidden error
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        self[1] = tag.Part(tag_name="div")


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the radio buttons"

        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if self.get_field_value('div_class'):
            self[1].update_attribs({"class":self.get_field_value('div_class')})
        checked = self.get_field_value('radio_checked')
        name = self.get_formname('radio_checked')
        label_class = self.get_field_value('label_class')
        self[1].insert_id()
        one_id = self[1].get_id()
        for index,val in enumerate(self.get_field_value('radio_values')):
            if self.get_field_value('inputdiv_class'):
                inputblock = tag.Part(tag_name="div", attribs={"class":self.get_field_value('inputdiv_class')})
            else:
                inputblock = tag.Part(tag_name="div")
            if checked and (checked == val):
                prt = tag.ClosedPart(tag_name="input",
                                            attribs ={"name":name,
                                                      "value":val,
                                                      "checked":"checked",
                                                      "type":"radio"})
                # ensure only one item checked
                checked = ''
            else:
                prt = tag.ClosedPart(tag_name="input",
                                            attribs ={"name":name,
                                                      "value":val,
                                                      "type":"radio"})
            prt.insert_id(one_id + '_' + str(index) + '_0')
            inputblock[0] = prt
            inputblock[1] = tag.Part(tag_name="label")
            if label_class:
                inputblock[1].attribs = {"class": label_class, 'for':inputblock[0].get_id()}
            else:
                inputblock[1].attribs = {'for':inputblock[0].get_id()}
            try:
                inputblock[1][0] = self.get_field_value('radio_text')[index]
            except IndexError:
                inputblock[1][0] = ''
            self[1].append(inputblock)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<div>  <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <div>  <!-- with class attribute set to div_class if a class is set -->
    <!-- Then for each radio field -->
    <div> <!-- with class set to inputdiv_class if a class is set -->
      <input type='radio' /> <!-- with name from 'radio_checked' widgfield and values from radio_values -->
      <label> <!-- with class set to label_class and content from radio_text -->
      </label>
    </div>
  </div>
</div>
"""


class RadioTable(Widget):
    """A div containing a table of radio controls with textblocks. Without a form or submit button, typically included within a form

       The first colum is the radio control with its associated label, the second column is either a textblock, or (if given)
       a string of text"""

    error_location = (0,0,0)

    arg_descriptions = {'table_class':FieldArg("cssclass", ''),
                        'header_class':FieldArg("cssclass", ''),
                        'even_class':FieldArg("cssclass", ''),
                        'odd_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", ''),
                        'title1':FieldArg("text"),
                        'title2':FieldArg("text"),
                        'radio_checked':FieldArg("text", valdt=True),
                        'contents':FieldArgTable(['text', 'text', 'textblock_ref', 'text'])
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        table_class: class of the table, if empty string, then no class will be applied
        header_class: class of the header row, if empty string, then no class will be applied
        even_class: class of even rows, if empty string, then no class will be applied
        odd_class: class of odd rows, if empty string, then no class will be applied
        error_class: css class applied to the normally hidden error paragraph
        title1: The header title over the first column
        title2: The header title over the second column
        radio_checked: value to be checked, this is the field returned
        contents: table of values, labels and textblock or text
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # hidden error
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        self[1] = tag.Part(tag_name='table')


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the table of radio buttons"
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")

        checked = self.get_field_value('radio_checked')
        name = self.get_formname('radio_checked')
        fieldtable = self.get_field_value("contents")
        if self.get_field_value('table_class'):
            self[1].attribs={"class":self.get_field_value('table_class')}
        # set table header with titles
        header = 0
        if self.get_field_value('title1') or self.get_field_value('title2'):
            header = 1
            if self.get_field_value('header_class'):
                self[1][0] = tag.Part(tag_name='tr', attribs={"class":self.get_field_value('header_class')})
            else:
                self[1][0] = tag.Part(tag_name='tr')
            self[1][0][0] = tag.Part(tag_name='th', text = self.get_field_value('title1'))
            self[1][0][1] = tag.Part(tag_name='th', text = self.get_field_value('title2'))
        # set even row colour
        if self.get_field_value('even_class'):
            even = self.get_field_value('even_class')
        else:
            even = ''
        # set odd row colour
        if self.get_field_value('odd_class'):
            odd = self.get_field_value('odd_class')
        else:
            odd = ''

        for index,val in enumerate(fieldtable):
            rownumber = index+header
            if even and (rownumber % 2) :
                self[1][rownumber] = tag.Part(tag_name="tr", attribs={"class":even})
            elif odd and not (rownumber % 2):
                self[1][rownumber] = tag.Part(tag_name='tr', attribs={"class":odd})
            else:
                self[1][rownumber] = tag.Part(tag_name='tr')
            # first col
            col0 = tag.Part(tag_name='td')
            if checked and (checked == val[0]):
                input_prt = tag.ClosedPart(tag_name="input",
                                            attribs ={"name":name,
                                                      "value":val[0],
                                                      "checked":"checked",
                                                      "type":"radio"})
                # ensure only one item checked
                checked = ''
            else:
                input_prt = tag.ClosedPart(tag_name="input",
                                            attribs ={"name":name,
                                                      "value":val[0],
                                                      "type":"radio"})
            input_label = tag.Part(tag_name='label')
            input_label[0] = input_prt
            input_label[1] = val[1]
            col0[0] = input_label
            # second col
            col1 = tag.Part(tag_name='td')
            if val[3]:
                # if text is given, this is set into the column
                col1[0] = val[3]
            else:
                # if no text, then the texblock is set
                col1[0] = val[2]
            # add the two columns to the row
            self[1][rownumber][0] = col0
            self[1][rownumber][1] = col1


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden paragraph, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <table>  <!-- with CSS class given by table_class -->
    <tr> <!-- with header class -->
      <th> <!-- title1 --> </th>
      <th> <!-- title2 --> </th>
    </tr>
    <!-- Then the following row repeated for each row given in the contents field -->
    <tr> <!-- with class  from even or odd classes -->
      <td>
        <label>
          <input type="radio" />
          <!-- with name from name of radio_checked field, and value from the contents -->
          <!-- label text from contents -->
        </label>
      </td>
      <td> <!-- textblock, found from row textbock_ref --> </td>
    </tr>
    <!-- rows repeated -->
  </table>
</div>"""
            


class TwoRadioOptions(Widget):
    """A two option radio control. Without a form or submit button, typically included within a form
       label appeares to the right of the buttons
        """
    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'label1':FieldArg("text"),
                        'div_class1':FieldArg("cssclass", ''),
                        'value1':FieldArg("text"),
                        'label2':FieldArg("text"),
                        'div_class2':FieldArg("cssclass", ''),
                        'value2':FieldArg("text"),
                        'radio_checked':FieldArg("text", valdt=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        div_class1: css class of div containing first radio box
        label1: text label to appear after the first radio button
        value1: the value returned if the first radio button is chosen
        div_class2: css class of div containing second radio box
        label2: text label to appear after the second radio button
        value2: the value returned if the second radio button is chosen
        radio_checked: value to be checked, this is the field returned
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name='div')
        self[0][0] = tag.Part(tag_name='label')
        self[1] = tag.Part(tag_name='div')
        self[1][0] = tag.Part(tag_name='label')


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the radio buttons"
        checked = self.get_field_value('radio_checked')
        name = self.get_formname('radio_checked')

        if self.get_field_value('div_class1'):
            self[0].update_attribs({"class":self.get_field_value('div_class1')})
        if self.get_field_value('div_class2'):
            self[1].update_attribs({"class":self.get_field_value('div_class2')})

        if checked and (checked == self.get_field_value('value1')):
            self[0][0][0] = tag.ClosedPart(tag_name="input",
                                     attribs ={"name":name,
                                     "value":self.get_field_value('value1'),
                                     "checked":"checked",
                                     "type":"radio"})
        else:
            self[0][0][0] = tag.ClosedPart(tag_name="input",
                                     attribs ={"name":name,
                                     "value":self.get_field_value('value1'),
                                     "type":"radio"})
        self[0][0][1] = self.get_field_value('label1')
        if checked and (checked == self.get_field_value('value2')):
            self[1][0][0] = tag.ClosedPart(tag_name="input",
                                     attribs ={"name":name,
                                     "value":self.get_field_value('value2'),
                                     "checked":"checked",
                                     "type":"radio"})
        else:
            self[1][0][0] = tag.ClosedPart(tag_name="input",
                                     attribs ={"name":name,
                                     "value":self.get_field_value('value2'),
                                     "type":"radio"})
        self[1][0][1] = self.get_field_value('label2')


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div>  <!-- with class attribute set to div_class1 if a class is set -->
    <label>
      <input type='radio' /> <!-- with name from 'radio_checked' widgfield and value from value1 -->
      <!-- text from label1 -->
    </label>
  </div>
  <div>  <!-- with class attribute set to div_class2 if a class is set -->
    <label>
      <input type='radio' /> <!-- with name from 'radio_checked' widgfield and value from value2 -->
      <!-- text from label2 -->
    </label>
  </div>
</div>
"""


class BooleanRadio(Widget):
    """A two option radio control for choosing True or False. Without a form or submit button, typically included within a form."""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'div_class1':FieldArg("cssclass", ''),
                        'left_label1':FieldArg("text", 'True'),
                        'left_class1':FieldArg("cssclass", ''),
                        'left_style1':FieldArg("cssstyle", ''),
                        'right_label1':FieldArg("text", ''),
                        'right_class1':FieldArg("cssclass", ''),
                        'right_style1':FieldArg("cssstyle", ''),
                        'div_class2':FieldArg("cssclass", ''),
                        'left_label2':FieldArg("text", 'False'),
                        'left_class2':FieldArg("cssclass", ''),
                        'left_style2':FieldArg("cssstyle", ''),
                        'right_label2':FieldArg("text", ''),
                        'right_class2':FieldArg("cssclass", ''),
                        'right_style2':FieldArg("cssstyle", ''),
                        'radio_checked':FieldArg("boolean", True, valdt=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        div_class1: class of the first div containing the first radio button
        left_label1: The text displayed to the left of the first radio button
        left_class1: The css class of the label to the left of the first radio button
        right_label1: The text displayed to the right of the first radio button
        right_class1: The css class of the label to the right of the first radio button
        div_class2: class of the second div containing the second radio button
        left_label2: The text displayed to the left of the second radio button
        left_class2: The css class of the label to the left of the second radio button
        right_label2: The text displayed to the right of the second radio button
        right_class2: The css class of the label to the right of the second radio button
        radio_checked: default value to be checked, True or False, field name is the field returned
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name='div')
        self[0][0] = tag.Part(tag_name="label", hide_if_empty=True)
        self[0][1] = tag.ClosedPart(tag_name="input")
        self[0][2] = tag.Part(tag_name="label", hide_if_empty=True)
        self[1] = tag.Part(tag_name='div')
        self[1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        self[1][1] = tag.ClosedPart(tag_name="input")
        self[1][2] = tag.Part(tag_name="label", hide_if_empty=True)


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the radio buttons"
        div1 = self[0]
        div2 = self[1]

        if self.get_field_value('div_class1'):
            div1.attribs = {"class": self.get_field_value('div_class1')}
        if self.get_field_value('div_class2'):
            div2.attribs = {"class": self.get_field_value('div_class2')}

        # labels of first div
        if self.get_field_value('left_class1'):
            div1[0].attribs = {"class": self.get_field_value('left_class1')}
        if self.get_field_value('left_style1'):
            div1[0].attribs = {"style": self.get_field_value('left_style1')}
        div1[0][0] = self.get_field_value('left_label1')

        if self.get_field_value('right_class1'):
            div1[2].attribs = {"class": self.get_field_value('right_class1')}
        if self.get_field_value('right_style1'):
            div1[2].attribs = {"style": self.get_field_value('right_style1')}
        div1[2][0] = self.get_field_value('right_label1')

        # labels of second div
        if self.get_field_value('left_class2'):
            div2[0].attribs = {"class": self.get_field_value('left_class2')}
        if self.get_field_value('left_style2'):
            div2[0].attribs = {"style": self.get_field_value('left_style2')}
        div2[0][0] = self.get_field_value('left_label2')

        if self.get_field_value('right_class2'):
            div2[2].attribs = {"class": self.get_field_value('right_class2')}
        if self.get_field_value('right_style2'):
            div2[2].attribs = {"style": self.get_field_value('right_style2')}
        div2[2][0] = self.get_field_value('right_label2')

        name = self.get_formname('radio_checked')
        if self.get_field_value('radio_checked'):
            # True chosen
            div1[1].attribs = {"type":"radio", 'name': name, "checked":"checked", "value":"True"}
            div2[1].attribs = {"type":"radio", 'name': name, "value":"False"}
        else:
            # False chosen
            div1[1].attribs = {"type":"radio", 'name': name, "value":"True"}
            div2[1].attribs = {"type":"radio", 'name': name, "checked":"checked", "value":"False"}

        # set the label 'for' attribute to the input field id
        div1[1].insert_id()
        div1[0].update_attribs({'for':div1[1].get_id()})
        div1[2].update_attribs({'for':div1[1].get_id()})
        div2[1].insert_id()
        div2[0].update_attribs({'for':div2[1].get_id()})
        div2[2].update_attribs({'for':div2[1].get_id()})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div>  <!-- with class attribute set to div_class1 if a class is set -->
      <label> <!-- with class set to left_class1 and content to left_label1 -->
      </label>
      <input type='radio' value='True' /> <!-- checked if radio_checked field is True -->
      <label> <!-- with class set to right_class1 and content to right_label1 -->
      </label>
  </div>
  <div>  <!-- with class attribute set to div_class2 if a class is set -->
      <label> <!-- with class set to left_class2 and content to left_label2 -->
      </label>
      <input type='radio' value='False' /> <!-- checked if radio_checked field is False -->
      <label> <!-- with class set to right_class2 and content to right_label2 -->
      </label>
  </div>
</div>"""








"""This module defines various widgets which provide information"""
import time

from string import Template

from .. import skiboot
from .. import tag
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class ServerTimeStamp(Widget):
    """A widget containing the current server time. consists of a span with time stamp"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'timestamp':FieldArg("text", '', jsonset=True),
                        'utc':FieldArg("boolean", True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        timestamp: normally empty and timestamp will automatically be displayed, or other text can be set here.
        utc: If True, time is utc, if False it is local time
        """
        Widget.__init__(self, name=name, tag_name="span", brief=brief, **field_args)
        self[0] = ""

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"
        # set timer or string
        if self.get_field_value("timestamp"):
            self[0] = self.get_field_value("timestamp")
        elif self.get_field_value("utc"):
            self[0] = time.strftime("%c", time.gmtime())
        else:
            self[0] = time.strftime("%c", time.localtime())

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- Normally timestamp -->
</span>
"""


class PageIdent(Widget):
    """A widget containing the given ident which is set within the text
         If no page ident is given, shows ident of the current page"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'page_ident':FieldArg("ident", ''),
                        'span_text':FieldArg("text", "", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        page_ident: The ident of a page or folder, converted to string
        span_text: if given, overrides the page_ident value
        """
        Widget.__init__(self, name=name, tag_name="span", brief=brief, **field_args)
        self[0] = ""

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"
        if self.get_field_value("span_text"):
            self[0] = self.get_field_value("span_text")
        elif self.get_field_value("page_ident"):
            self[0] = self.get_field_value("page_ident").to_comma_str()
        else:
            self[0] = page.ident.to_comma_str()


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- string value of the page_ident, or span_text -->
</span>
"""


class PageName(Widget):
    """A widget containing the given page name which is set within the text
         If no page ident is given, shows name of the current page"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'page_ident':FieldArg("ident", ''),
                        'span_text':FieldArg("text", "", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        page_ident: The ident of a page or folder
        span_text: if given, overrides the page name value
        """
        Widget.__init__(self, name=name, tag_name="span", brief=brief, **field_args)
        self[0] = ""

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"
        if self.get_field_value("span_text"):
            self[0] = self.get_field_value("span_text")
        elif self.get_field_value("page_ident"):
            requested_page = skiboot.get_item(self.get_field_value("page_ident"))
            if requested_page is None:
                self[0] = "Unknown page"
            else:
                self[0] = requested_page.name
        else:
            self[0] = page.name


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- The page name, or span_text -->
</span>
"""


class PageDescription(Widget):
    """A widget containing the page brief of the given ident which is set within the text
         If no page ident is given, shows brief of the current page"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'page_ident':FieldArg("ident", ''),
                        'span_text':FieldArg("text", "", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        page_ident: The ident of a page or folder
        span_text: if given, overrides the page description value
        """
        Widget.__init__(self, name=name, tag_name="span", brief=brief, **field_args)
        self[0] = ""

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"
        if self.get_field_value("span_text"):
            self[0] = self.get_field_value("span_text")
        elif self.get_field_value("page_ident"):
            requested_page = skiboot.get_item(self.get_field_value("page_ident"))
            if requested_page is None:
                self[0] = "Unknown page"
            else:
                self[0] = requested_page.brief
        else:
            self[0] = page.brief

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- The page description -->
</span>
"""


class ProjectName(Widget):
    """A span showing the name of the project the page is in"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {}

    def __init__(self, name=None, brief='', **field_args):
        """
        Shows project name
        """
        Widget.__init__(self, name=name, tag_name="span", brief=brief, **field_args)
        self[0] = ""

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"        
        self[0] = page.ident.proj

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- the project name -->
</span>
"""


class Version(Widget):
    """A span showing the project version of the page it is in"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {}

    def __init__(self, name=None, brief='', **field_args):
        """
        Shows project version
        """
        Widget.__init__(self, name=name, tag_name="span", brief=brief, **field_args)
        self[0] = ""

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"        
        self[0] = page.project.version

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- string of the project version -->
</span>
"""


class SkipoleVersion(Widget):
    """A span showing this version of the skipole framework"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {}

    def __init__(self, name=None, brief='', **field_args):
        """
        Shows skipole version
        """
        Widget.__init__(self, name=name, tag_name="span", brief=brief, **field_args)
        self[0] = ""

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"        
        self[0] = skiboot.version()

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- string of the skipole framework version -->
</span>
"""


class Redirector(Widget):
    """A widget containing javascript which redirects the page to the url
       Once the page is loaded, the redirection occurs, so normally
       this is the only widget on the page
       A textblock is displayed with the link to the url if the client has javascript disabled"""

    arg_descriptions = {'url':FieldArg("text", ''),
                        'textblock_ref':FieldArg("textblock_ref", ""),
                        'text_refnotfound':FieldArg("text", "If this page does not redirect automatically, follow this link:"),
                        'text_replaceblock':FieldArg("text", "If this page does not redirect automatically, follow this link:"),
                        'linebreaks':FieldArg("boolean", True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        url: the url to redirect to
        textblock_ref: The reference of the TextBlock appearing in the paragraph
        text_refnotfound: text to appear if the textblock is not found
        text_replaceblock: text set here will replace the textblock
        linebreaks: Set True if linebreaks in the text are to be shown as html breaks
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)

        self[0] = tag.Part(tag_name='script', attribs={"type":"text/javascript"})
        self[1] = tag.Part(tag_name='p')
        self._url = ''


    def _error_build(self, message):
        """Called if an error is raised"""
        if message:
            self._url = message


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the widget"
        if self.error_status and self._url:
            url = self._url
        else:
            url = self.get_field_value("url")

        self[0][0] = "window.location.replace(\"%s\");" % (url,)

        linebreaks = bool(self.get_field_value('linebreaks'))

        if self.get_field_value('text_replaceblock'):
            # no textblock, just the replacement text
            tblock = self.get_field_value('text_replaceblock')
            if not linebreaks:
                self[1].linebreaks = False
        else:
            # define the textblock
            tblock = self.get_field_value("textblock_ref")
            tblock.failmessage = self.get_field_value('text_refnotfound')
            tblock.linebreaks = linebreaks
            tblock.proj_ident = page.proj_ident

        self[1][0] = tblock
        self[1][1] = tag.ClosedPart(tag_name='br')
        self[1][2] = tag.Part(tag_name='a', attribs={"href":url}, text=url)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <!-- url is either the url argument, or is the error message if an error is raised on the widget -->
  <script type="text/javascript">
     <!-- window.location.replace("url"); -->
  </script>
  <p>
    <!-- text_replaceblock or the textblock text here -->
    <br />
    <a href="url"> 
        <!-- url --> 
    </a>
  </p>
</div>
"""






"""Functions that access the ski framework"""

import os

from ..ski.excepts import FailPage, GoTo, ValidateError, ServerError

from ..ski import skiboot, tag

from .info_tuple import ProjectInfo, ItemInfo, PartInfo, PageInfo, FolderInfo, WidgetInfo, VersionInfo


def project_loaded(project, error_if_not=True):
    """Returns True if the project is the root or a sub project.

     If the given project is not one of these, either raises a ServerError exception (if error_if_not is True) or returns False.
"""
    if isinstance(project, str) and skiboot.is_project(project):
        return True
    if error_if_not:
        raise ServerError(message="The project has not been loaded")
    return False


def get_proj(project):
    """Returns the project instance, being the object used internally by the skipole framework"""
    project_loaded(project)
    return skiboot.getproject(project)


def get_proj_page(project, pagenumber, pchange=None):
    """Returns (project_object, page_object), being objects used internally by the skipole framework
       The page object returned is a deepcopy of the page in the project""" 

    project_loaded(project)
    proj = skiboot.getproject(project)

    # get a copy of the page
    if not isinstance(pagenumber, int):
        raise ServerError(message="Given pagenumber is not an integer")
    ident = skiboot.Ident.to_ident((project, pagenumber))
    if ident is None:
        raise ServerError(message="Invalid project, pagenumber")
    page = skiboot.from_ident(ident, project)
    if page is None:
        raise ServerError(message="Invalid Page - pagenumber not found in project")
    if page.page_type == 'Folder':
        raise ServerError(message = "Invalid page - requested item is a Folder")
    # if pchange is given, test it is equal to page.change
    if (pchange is not None) and (page.change != pchange):
        raise ServerError(message="The page has been changed prior to this submission, someone else may be editing this project")
    return proj, page


def get_proj_section(project, section_name, schange=None):
    """Returns (project_object, section_object), being objects used internally by the skipole framework
       The section object returned is a deepcopy of the section in the project""" 

    project_loaded(project)
    proj = skiboot.getproject(project)

    # get a copy of the section
    if not isinstance(section_name, str):
        raise ServerError(message="Given section_name is invalid")
    section = proj.section(section_name, makecopy=True)
    if section is None:
        raise ServerError(message="Given section_name is invalid")
    # if schange is given, test it is equal to section.change
    if (schange is not None) and (section.change != schange):
        raise ServerError(message="The section has been changed prior to this submission, someone else may be editing this project")
    return proj, section


def versions(project):
    """Returns a namedtuple with contents
       skipole, project - these being the skipole version, and the project version
    """
    project_loaded(project)
    proj = skiboot.getproject(project)
    return VersionInfo(skiboot.version(), proj.version)


def project_info(project):
    """Returns a namedtuple with contents
       project, version, brief, path, default_language, subprojects, json_path, tar_path, main_path, static_path, data_path
       where subprojects is an ordered dictionary of projectname:path
    """
    project_loaded(project)
    proj = skiboot.getproject(project)
    return ProjectInfo(
                   project,
                   proj.version,
                   proj.brief,
                   proj.url,
                   proj.default_language,
                   proj.subproject_paths,
                   skiboot.project_json(project),
                   skiboot.project_main(project),
                   skiboot.projectstatic(project),
                   skiboot.projectdata(project)
                   )


def projectURLpaths():
    """Returns a dictionary of project name : project path

       This function takes no arguments and returns a dictionary of project names as keys with the URL paths as values.
       The projects being the root project and any sub-projects attached to it.
       Please note, if this project is being administered using 'skiadmin' - then the project 'skiadmin' will be visible in this dictionary,
       however when the project is running without skiadmin, this project will no longer be present.

      You should expect to see the root project itself in the dictionary, and at least the 'skis' project which is normally attached as a sub-project.
"""
    all_projects = skiboot.project_register()
    return {proj_ident:proj.url for proj_ident, proj in all_projects.items()}


def get_root():
    "Takes no arguments and returns the project name of the current root project."
    return skiboot.root_project().proj_ident


def lib_list():
    "Returns list of library labels"
    return skiboot.lib_list()[:]


def sys_list():
    "Returns list of system page labels"
    return skiboot.sys_list()[:]


def get_debug():
    "Returns the debug mode"
    return skiboot.get_debug()

def set_debug(mode):
    "Sets debug mode"
    skiboot.set_debug(mode)


def get_proj_data(project, key):
    "Returns the value from the proj_data dictionary, or None if not found"
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        return
    return proj.proj_data.get(key)

def set_proj_data(project, key, value):
    "Sets the key, value in the proj_data dictionary"
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        return
    proj.proj_data[key] = value


def set_proj_default_language(project, default_language):
    "Sets the default language of the project"
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        return
    proj.default_language = default_language


def set_proj_brief(project, brief):
    "Sets the brief description of the project"
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        return
    proj.brief = brief


def set_proj_version(project, version):
    "Sets the version of the project"
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        return
    proj.version = version


def get_projectfiles_dir(project=None):
    """Returns the projectfiles directory for the given project

    Static files, and control files (such as the json files defining your project) are saved outside
    the python package structure under a 'projectfiles' directory.
    If project not given, returns the projectfiles directory path for the root project"""
    return skiboot.projectfiles(project)


def get_projectdir(project=None):
    """Returns the projectfiles/project directory for the given project

    Static files, and control files (such as the json files defining your project) are saved outside
    the python package structure under a 'projectfiles/project' directory.
    If project not given, returns the projectfiles/project path for the root project"""
    return skiboot.projectdir(project)


def get_projectcode_dir(project=None):
    """Returns the directory where your project code is saved
    If project not given, returns the projectcode directory path for the root project"""
    return skiboot.projectcode(project)

def get_projectstatic_dir(project=None):
    """Returns the directory where your project static files are saved
    If project not given, returns the projectstatic directory path for the root project"""
    return skiboot.projectstatic(project)


def next_ident_number(project):
    "Returns next ident number in the project by incrementing highest existing number"
    proj = skiboot.getproject(project)
    if proj is None:
        raise ServerError(message="Project not loaded")
    return proj.next_ident().num


def get_textblock_text(textref, lang, project=None):
    """This function returns the textblock text, given a textblock reference string,

       lang is the language tuple of (preferred language, default language).
       If text for a given language is not found, the function makes an effort to get text by trying the default language.

       If project is not given assumes the root project, if given, project must exist as either the root,
       or a sub project of the root.
       If no textblock is found, returns None."""
    if project is None:
        project = skiboot.project_ident()
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        return
    return proj.textblocks.get_text(textref, lang)


def get_accesstextblocks(project=None):
    """This function returns the instance of the AccessTextBlocks class used by the project to get TextBlock text.

       If project is None, assumes the root project, if project is given; returns the AccessTextBlocks instance
       of the project. If given, project must exist as either the root, or a sub project of the root."""
    if project is None:
        project = skiboot.project_ident()
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        return
    return proj.textblocks


def get_itemnumber(project, item):
    """Returns itemnumber if item is found in the project, where item is either
    a project,number tuple
    a number
    a label
    a path
    If not found returns None"""
    # return None if invalid project
    if not project_loaded(project, error_if_not=False):
        return
    proj = skiboot.getproject(project)
    if proj is None:
        return
    # check if item is a path
    if isinstance(item, str) and '/' in item:
        path = item.strip()
        path = path.strip("/")
        if path:
            path = "/" + path + "/"
        else:
            path = "/"
        ident = proj.root.ident_from_path(path)
    else:
        ident = skiboot.find_ident(item, project)
    if ident is None:
        return
    if ident.proj != project:
        return
    return ident.num



def item_info(project, itemnumber):
    """Returns None if page or folder not found, otherwise returns a namedtuple of item information.

       The project must exist as either the root, or a sub project of the root, itemnumber is the ident number of a page or folder.

       The function returns None, if no page or folder is found, otherwise returns a namedtuple with contents:
       project, project_version, itemnumber, item_type, name, brief, path, label_list, change, parentfolder_number, restricted.

       The values are:
       'project' - the project name from the argument.
       'project_version' is the project version string.
       'itemnumber' is the itemnumber from the argument.
       item_type' - one of 'TemplatePage', 'Folder', 'JSON', 'CSS', 'RespondPage', 'FilePage', 'SVG'.
       'name' holds the page or folder name.
       'brief' holds the brief description of the item.
       'path' is the url path.
       'label_list' is a list of labels (as there could be more than one) pointing to this item.
       'change' is a uuid of the item, changed whenever the item is altered by this admin site.
       'parentfolder_number' is the integer number of the parentfolder or None if this is the root folder.
       'restricted' is True if access is restricted (cannot be called by URL), False otherwise.
    """
    # raise error if invalid project
    project_loaded(project)
    if not isinstance(itemnumber, int):
         raise ServerError(message="Given itemnumber is not an integer")
    ident = skiboot.Ident.to_ident((project, itemnumber))
    if ident is None:
        return
    info = skiboot.item_info(ident)
    if not info:
        return
    if info.parentfolder_ident:
        parentfolder_number = info.parentfolder_ident.num
    else:
        parentfolder_number = None

    return ItemInfo(
                   project,
                   info.project_version,
                   itemnumber,
                   info.item_type,
                   info.name,
                   info.brief,
                   info.path,
                   info.label_list,
                   info.change,
                   parentfolder_number,
                   info.restricted
                   )


def part_info(project, pagenumber, section_name, location):
    """Returns a PartInfo named tuple, giving the information about a part at a location in a page or section.

       The project must exist as either the root, or a sub project of the root,
       pagenumber and section_name are mutually exclusive, one must be None.

       location is a tuple or list consisting of three items:
       a string (such as 'head' or section name or widget name)
       a container integer, such as 0 for widget container 0, or None if not in container
       a tuple or list of location integers

       Returns None if part not found, otherwise returns a namedtuple with items
       project, pagenumber, page_part, section_name, name, location, part_type, brief.

       page_part is one of 'head', 'body', 'svg' or None if the part is in a section,
       name is the widgets name if it is a widget.
    """

    # part is either in a page or a section
    if (pagenumber is None) and (section_name is None):
        raise ServerError("Page and section both missing")

    page_part = None

    if section_name:
        if pagenumber is not None:
            raise ServerError("Part cannot be in both a page and a section")
        # item is in a section
        proj, section = get_proj_section(project, section_name)
        part = section.location_item(location)
    else:
        # item is in a page
        proj, page = get_proj_page(project, pagenumber)
        part = page.location_item(location)
        page_part = location[0]
        if location[1] is not None:
            # item is in a container in a widget, so location[0] will be the parent widget name
            widget = page.widgets[location[0]]
            if widget is not None:
               ident_top = widget.ident_string.split("-", 1)
               # ident_top[0] will be of the form proj_pagenum_head
               page_part = ident_top[0].split("_")[2]

    if part is None:
        raise ServerError("Part not found")

    if hasattr(part, '__class__'):
        part_type = part.__class__.__name__
    else:
        part_type = None

    if hasattr(part, 'name'):
        name = part.name
    else:
        name = None

    if hasattr(part, 'brief'):
        brief = part.brief
    else:
        brief = None

    return PartInfo(project, pagenumber, page_part, section_name, name, location, part_type, brief)


def part_contents(project, pagenumber, section_name, location):
    "If the given part is a Part or Section, returns a list of PartInfo tuples, one for each content"
    # part is either in a page or a section
    if (pagenumber is None) and (section_name is None):
        raise ServerError("Page and section both missing")

    page_part = None

    if section_name:
        if pagenumber is not None:
            raise ServerError("Part cannot be in both a page and a section")
        # item is in a section
        proj, section = get_proj_section(project, section_name)
        part = section.location_item(location)
    else:
        # item is in a page
        proj, page = get_proj_page(project, pagenumber)
        part = page.location_item(location)
        if location[1] is not None:
            # item is in a container in a widget, so location[0] will be the parent widget name
            widget = page.widgets[location[0]]
            if widget is not None:
               ident_top = widget.ident_string.split("-", 1)
               # ident_top[0] will be of the form proj_pagenum_head
               page_part = ident_top[0].split("_")[2]

    if part is None:
        return
    if hasattr(part, '__class__'):
        part_type = part.__class__.__name__
    if (part_type != "Part") and (part_type != "Section"):
        return

    subpart_list = []

    index = 0
    for subpart in part:
        if hasattr(subpart, '__class__'):
            subpart_type = subpart.__class__.__name__
        else:
            subpart_type = None

        if hasattr(subpart, 'name'):
            name = subpart.name
        else:
            name = None

        if hasattr(subpart, 'brief'):
            brief = subpart.brief
        else:
            brief = None

        location_list = list(location[2])

        sublocation = (location[0], location[1], location_list+[index])

        sub_tuple = PartInfo(project, pagenumber, page_part, section_name, name, sublocation, subpart_type, brief)
        subpart_list.append(sub_tuple)
        index += 1
    return subpart_list


def widget_info(project, pagenumber, section_name, widget_name):
    """returns None if widget not found, otherwise returns a namedtuple with items
       project, pagenumber, section_name, name, location, containers, display_errors, brief
       where containers is the number of containers this widget has, zero if it has none,
       and display_errors is True if this widget can display an error"""
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    location = None
    # get the widget
    if section_name:
        section = proj.section(section_name, makecopy=False)
        if section is None:
            return
        if widget_name not in section.widgets:
            return
        widget = section.widgets[widget_name]
        if widget is None:
            return
        parent, container = widget.get_parent_widget(section)
        ident_items = widget.ident_string.split('-', 1)
        ident_integers = ident_items[1].split('-')
        if parent is None:
            location = (section_name, None, tuple(int(i) for i in ident_integers))
    elif pagenumber:
        page = proj.get_item(pagenumber)
        if widget_name not in page.widgets:
            return
        widget = page.widgets[widget_name]
        if widget is None:
            return
        parent, container = widget.get_parent_widget(page)
        ident_items = widget.ident_string.split('-', 1)
        ident_integers = ident_items[1].split('-')
        ident_locs = ident_items[0].split('_')
        # ident_locs are project, pagenumber, (head,body or svg)
        if parent is None:
            location = (ident_locs[2], None, tuple(int(i) for i in ident_integers))
    else:
        return
    if location is None:
        parent_ident_items = parent.ident_string.split('-', 1)
        parent_ident_integers = parent_ident_items[1].split('-')
        length_of_container_location = len(parent_ident_integers) + len(parent.get_container_loc(container))
         # widget location integers, are the integers beyond the parent + container integers
        location_integers = ident_integers[length_of_container_location:]
        location = (parent.name, container, tuple(location_integers))
    return WidgetInfo(project, pagenumber, section_name, widget_name, location, widget.len_containers(), widget.display_errors, widget.brief)



def ident_exists(project, itemnumber):
    "Return True if a page or folder with this itemnumber exists, False otherwise."
    return skiboot.ident_exists((project, itemnumber))


def ident_numbers(project=None):
    "Returns a list of the project ident numbers"
    num_list = [0]
    if project is None:
        project = skiboot.project_ident()
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        return []
    return proj.ident_numbers


def labels(project=None):
    """Returns a dictionary of labels to page, folder tuple idents or urls

       If project is not given assumes the root project, if given, project must exist as either the root, or a sub project of the root.
"""
    if project is None:
        project = skiboot.project_ident()
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        return {}
    return proj.labels()


def ident_from_label(project, label):
    """Returns (itemproject, itemnumber) or url if found, otherwise None"""
    item = skiboot.find_ident_or_url(label, project)
    if isinstance(item, skiboot.Ident):
        return item.proj, item.num
    return item



def set_label(project, label, target):
    """Sets a label and target, where target is a page ident or url"""
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        raise ServerError("Project not recognised")
    proj.set_special_page(label, target)


def delete_label(project, label):
    """Deletes a label"""
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        raise ServerError("Project not recognised")
    proj.delete_special_page(label)


def pages(project, foldernumber):
    """Returns generator of PageInfo named tuples, one for each page in the folder

       The project must exist as either the root, or a sub project of the root, foldernumber is the ident number of a folder.

       The generator raises a ServerError if no folder is found, otherwise the generator yields a namedtuple describing each page within the folder over each iterated pass.

       The named tuple has contents:
       'name', 'number', 'restricted', 'brief', 'item_type', 'responder', 'enable_cache', 'change', 'parentfolder_number'

       The values are:
       'name' holds the page name.
       'number' is the ident page number.
       'restricted' is True if access is restricted (cannot be called by URL), False otherwise.
       'brief' holds the brief description of the page.
       'item_type' - one of 'TemplatePage', 'JSON', 'CSS', 'RespondPage', 'FilePage', 'SVG'.
       'responder' - responder Class name, such as 'SubmitData' etc., or empty string if the page is not a responder.
       'enable_cache' - True if client caching is enabled on the page, False otherwise.
       'change' - A UUID string changed whenever the page is edited
       'parentfolder_number' - The uident number of the folder containing this page
"""

    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    folder_ident = skiboot.Ident(project, foldernumber)
    folder = proj.get_item(folder_ident)
    if folder is None:
        raise ServerError("Folder not recognised")
    for page_ident in folder.page_idents():
        page = proj.get_item(page_ident)
        if page.page_type == 'RespondPage':
            responder = page.responder.__class__.__name__
        else:
            responder = ''
        if hasattr(page, 'enable_cache'):
            enable_cache = page.enable_cache
        else:
            enable_cache = False
        if page.parentfolder_ident:
            parentfolder_number = page.parentfolder_ident.num
        else:
            parentfolder_number = None
        yield PageInfo(page.name, page.ident.num, page.restricted, page.brief, page.page_type, responder, enable_cache, page.change, parentfolder_number)


def page_info(project, pagenumber):
    """Returns PageInfo named tuple for page

       The project must exist as either the root, or a sub project of the root, pagenumber is the ident number of a page.

       The function raises a ServerError, if no page is found, otherwise returns a namedtuple with contents:
       'name', 'number', 'restricted', 'brief', 'item_type', 'responder', 'enable_cache', 'change', 'parentfolder_number'

       The values are:
       'name' holds the page name.
       'number' is the page number from the argument.
       'restricted' is True if access is restricted (cannot be called by URL), False otherwise.
       'brief' holds the brief description of the page.
       'item_type' - one of 'TemplatePage', 'JSON', 'CSS', 'RespondPage', 'FilePage', 'SVG'.
       'responder' - responder Class name, such as 'SubmitData' etc., or empty string if the page is not a responder.
       'enable_cache' - True if client caching is enabled on the page, False otherwise.
       'change' - A UUID string changed whenever the page is edited
       'parentfolder_number' - The uident number of the folder containing this page
"""
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    page_ident = skiboot.Ident(project, pagenumber)
    page = proj.get_item(page_ident)
    if page is None:
        raise ServerError("Page not recognised")
    if page.page_type == 'Folder':
        raise ServerError("Item is a Folder, not a page")
    if page.page_type == 'RespondPage':
        responder = page.responder.__class__.__name__
    else:
        responder = ''
    if hasattr(page, 'enable_cache'):
        enable_cache = page.enable_cache
    else:
        enable_cache = False
    if page.parentfolder_ident:
        parentfolder_number = page.parentfolder_ident.num
    else:
        parentfolder_number = None
    return PageInfo(page.name, page.ident.num, page.restricted, page.brief, page.page_type, responder, enable_cache, page.change, parentfolder_number)


def folders(project, foldernumber):
    """Returns generator of FolderInfo named tuples, one for each sub folder in the folder

    The project must exist as either the root, or a sub project of the root, foldernumber is the ident number of a folder.

    The generator raises a ServerError if no folder is found, otherwise the generator yields a namedtuple describing each sub folder within the folder over each iterated pass.

    The named tuple has contents:
    'name', 'number', 'restricted', 'brief', 'contains_pages', 'contains_folders', 'change'.

    The values are:
    'name' holds the sub folder name.
    'number' is the sub folder number.
    'restricted' is True if access is restricted (cannot be called by URL), False otherwise.
    'brief' holds the brief description of the sub folder.
    'contains_pages' - True if the sub folder contains pages, False otherwise.
    'contains_folders' - True if the sub folder contains pages, False otherwise.
    'change' - A UUID string, changed whenever the folder is changed.
"""
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    folder_ident = skiboot.Ident(project, foldernumber)
    folder = proj.get_item(folder_ident)
    if folder is None:
        raise ServerError("Folder not recognised")
    for subfolder_ident in folder.folder_idents():
        subfolder = proj.get_item(subfolder_ident)
        yield FolderInfo(subfolder.name, subfolder.ident.num, subfolder.restricted, subfolder.brief, bool(subfolder.pages), bool(subfolder.folders), subfolder.change)


def folder_info(project, foldernumber):
    """Returns FolderInfo named tuple for folder

    The project must exist as either the root, or a sub project of the root, foldernumber is the ident number of a folder.

    The function raises a ServerError, if no folder is found, otherwise returns a namedtuple with contents:
    'name', 'number', 'restricted', 'brief', 'contains_pages', 'contains_folders', 'change'.

    The values are:
    'name' holds the folder name.
    'number' is the folder number from the argument.
    'restricted' is True if access is restricted (cannot be called by URL), False otherwise.
    'brief' holds the brief description of the folder.
    'contains_pages' - True if the folder contains pages, False otherwise.
    'contains_folders' - True if the folder contains pages, False otherwise.
    'change' - A UUID string, changed whenever the folder is changed
"""
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    folder_ident = skiboot.Ident(project, foldernumber)
    folder = proj.get_item(folder_ident)
    if folder is None:
        raise ServerError("Folder not recognised")
    if folder.page_type != 'Folder':
        raise ServerError("Item is not a Folder")
    return FolderInfo(folder.name, folder.ident.num, folder.restricted, folder.brief, bool(folder.pages), bool(folder.folders), folder.change)


def folder_page_names(project, foldernumber):
    """Returns tuple of (default_page_name, [page names,....]) where default_page_name is None if no default page has been set
       and the page names list is empty if no pages are present. The page names list is sorted."""
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    folder_ident = skiboot.Ident(project, foldernumber)
    folder = proj.get_item(folder_ident)
    if folder is None:
        raise ServerError("Folder not recognised")
    # folder.pages is a dictionary of page.name:page.ident
    if not folder.pages:
        return None, []
    page_names = list(folder.pages.keys())
    page_names.sort()
    if folder.default_page_name:
        return folder.default_page_name, page_names
    return None, page_names


def parent_list(project, itemnumber):
    """Returns list of [(name,number),...] of parents, starting with ('root',0) ending with name and number of item"""
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    item_ident = skiboot.Ident(project, itemnumber)
    item = proj.get_item(item_ident)
    if item is None:
        raise ServerError("Item not recognised")
    return item.parent_list()


def page_path(project, item):
    "Returns a url path given a page or folder number or label, if not found, return None"
    # raise error if invalid project
    project_loaded(project)
    return skiboot.get_url(item, project)


def insert_item_in_page(project, pagenumber, pchange, location, item):
    "Insert the item in the page at location, return the new pchange value and new location"
    proj, page = get_proj_page(project, pagenumber, pchange)
    if (page.page_type != 'TemplatePage') and (page.page_type != 'SVG'):
        raise ServerError(message="The page must be a Template or SVG page")
    if hasattr(item, 'name') and item.name:
        name = item.name
        if name in page.widgets:
            raise ServerError(message="The name clashes with a widget name already in the page")
        if name in page.section_places:
            raise ServerError(message="This name clashes with a section alias within this page")
    if hasattr(item, 'placename'):
        name = item.placename
        if name in page.widgets:
            raise ServerError(message="The alias clashes with a widget name already in the page")
        if name in page.section_places:
            raise ServerError(message="This alias clashes with another section alias within this page")

    location_string, container, location_integers = location
    location_integers = [int(i) for i in location[2]]
    # location string is either a widget name, or body, head, or svg
    # if a widget_name, container must be given
    # get the part at the current location
    if container is None:
        # not in a widget
        parent_widget = None
        part = page.location_item((location_string, None, location_integers))
    else:
        # so item is in a widget, location_string is the widget name
        parent_widget = page.widgets[location_string]
        part = parent_widget.get_from_container(container, location_integers)


    # If this item is to be placed inside a parent widget container
    if (parent_widget is not None) and (parent_widget.is_container_empty(container)):
        # item is to be set as the first item in a widget container
        new_location = (location_string, container, (0,))
        parent_widget.set_in_container(container, (0,), item)
    elif isinstance(part, tag.Part) and (not hasattr(part, "arg_descriptions")): # not Closed Part and not a widget
        # insert at position 0 inside the part
        new_location = (location_string, container, tuple(location_integers + [0]))
        part.insert(0,item)
    elif (parent_widget is not None) and (len(location_integers) == 1):
        # part is inside a container with parent being the containing div
        # so append after the part by inserting at the right place in the container
        position = location_integers[0] + 1
        new_location = (location_string, container, (position,))
        parent_widget.insert_into_container(container, position, item)
    else:
        # do an append, rather than an insert
        # get parent part
        loc_integers = location_integers[:-1]
        if (location_string == 'head') or (location_string == 'body') or (location_string == 'svg'):
            parent_part = page.location_item((location_string, None, loc_integers))
        else:
            # parent_widget is the containing widget 
            parent_part = parent_widget.get_from_container(container, loc_integers)

        # find location digit
        loc = location_integers[-1] + 1
        new_location = (location_string, container, tuple(location_integers[:-1] + [loc]))
        # insert item at loc in parent_part
        parent_part.insert(loc,item)

    # save the altered page and return the change uuid, and new location
    return proj.save_page(page), new_location


def insert_item_in_section(project, section_name, schange, location, item):
    "Insert the item in the section at location, return the new schange value and new location"
    proj, section = get_proj_section(project, section_name, schange)
    if hasattr(item, 'name') and item.name:
        name = item.name
        if name in section.widgets:
            raise ServerError(message="The name clashes with a widget name already in the section")
    location_string, container, location_integers = location
    location_integers = [int(i) for i in location[2]]
    # location string is either a widget name, or the section name
    # if a widget_name, container must be given
    # get the part at the current location
    if container is None:
        # not in a widget
        parent_widget = None
        if location_string != section_name:
            raise ServerError(message="Invalid location, section name not equal to location string")
        if location_integers:
            part = section.get_location_value(location_integers)
        else:
            part = section
    else:
        # so item is in a widget, location_string is the widget name
        parent_widget = section.widgets[location_string]
        part = parent_widget.get_from_container(container, location_integers)


    # If this item is to be placed inside a parent widget container
    if (parent_widget is not None) and (parent_widget.is_container_empty(container)):
        # item is to be set as the first item in a widget container
        new_location = (location_string, container, (0,))
        parent_widget.set_in_container(container, (0,), item)
    elif isinstance(part, tag.Part) and (not hasattr(part, "arg_descriptions")): # not Closed Part and not a widget
        # insert at position 0 inside the part
        new_location = (location_string, container, tuple(location_integers + [0]))
        part.insert(0,item)
    elif (parent_widget is not None) and (len(location_integers) == 1):
        # part is inside a container with parent being the containing div
        # so append after the part by inserting at the right place in the container
        position = location_integers[0] + 1
        new_location = (location_string, container, (position,))
        parent_widget.insert_into_container(container, position, item)

    else:
        # do an append, rather than an insert
        # get parent part
        loc_integers = location_integers[:-1]
        if parent_widget is None:
            # part not in a widget
            if location_string == section_name:
                if loc_integers:
                    parent_part = section.get_location_value(loc_integers)
                else:
                    parent_part = section
        else:
            parent_part = parent_widget.get_from_container(container, loc_integers)

        # find location digit
        loc = location_integers[-1] + 1
        new_location = (location_string, container, tuple(location_integers[:-1] + [loc]))
        # insert item at loc in parent_part
        parent_part.insert(loc,item)

    # save the altered section, and return the change uuid, and new location
    return proj.add_section(section_name, section), new_location



def del_location_in_page(project, pagenumber, pchange, location):
    "Deletes the item at the given location in the page"
    # raise error if invalid project
    proj, page = get_proj_page(project, pagenumber, pchange)
    if (page.page_type != 'TemplatePage') and (page.page_type != 'SVG'):
        raise ServerError(message = "Invalid page")

    location_string, container, location_integers = location

    # location string is either a widget name, or body, head, or svg
    # if a widget_name, container must be given

    if container is None:
        # not in a widget
        if location_string == 'body':
            top = page.body
        elif location_string == 'head':
            top = page.head
        elif location_string == 'svg':
            top = page.svg
        else:
            raise ServerError(message="Given location is invalid")
        # remove the item
        try:
            top.del_location_value(location_integers)
        except Exception:
            raise ServerError(message="Unable to delete item")
        # And save this page copy to the project
        return proj.save_page(page)

    # so item is in a widget, location_string is the widget name
    widget = page.widgets[location_string]
    ident_string = widget.ident_string

    # ident_string is of the form; project_pageidentnumber_body-x-y

    splitstring = ident_string.split("_")
    splitloc = splitstring[2].split("-")
    loc_top = splitloc[0]
    widg_ints = [ int(i) for i in splitloc[1:] ]

    widg_container_ints = list(widget.get_container_loc(container))

    item_location_ints = widg_ints + widg_container_ints + list(location_integers)

    if loc_top == 'body':
        top = page.body
    elif loc_top == 'head':
        top = page.head
    elif loc_top == 'svg':
        top = page.svg
    else:
        raise ServerError(message="Given location is invalid")
    # remove the item
    try:
        top.del_location_value(item_location_ints)
    except Exception:
        raise ServerError(message="Unable to delete item")
    # And save this page copy to the project
    return proj.save_page(page)


def del_location_in_section(project, section_name, schange, location):
    "Deletes the item at the given location"
    # raise error if invalid project
    proj, section = get_proj_section(project, section_name, schange)

    location_string, container, location_integers = location

    if container is None:
        if location_string != section_name:
            raise ServerError(message="Unable to delete item")
        # remove the item
        try:
            section.del_location_value(location_integers)
        except Exception:
            raise ServerError(message="Unable to delete item")
        # And save this section copy to the project
        return proj.add_section(section_name, section)

    # so item is in a widget, location_string is the widget name
    widget = section.widgets[location_string]
    ident_string = widget.ident_string

    # ident_string is sectionname-x-y

    splitstring = ident_string.split("-")
    widg_ints = [ int(i) for i in splitstring[1:] ]

    widg_container_ints = list(widget.get_container_loc(container))

    item_location_ints = widg_ints + widg_container_ints + list(location_integers)

    # remove the item
    try:
        section.del_location_value(item_location_ints)
    except Exception:
        raise ServerError(message="Unable to delete item")
    # And save this section copy to the project
    return proj.add_section(section_name, section)



def set_item_in_page(project, pagenumber, pchange, location, item):
    "Sets an item in the location"

    proj, page = get_proj_page(project, pagenumber, pchange)
    if (page.page_type != 'TemplatePage') and (page.page_type != 'SVG'):
        raise ServerError(message = "Invalid page")

    location_string, container, location_list = location

    # set item in a page
    if (location_string == 'head'):
        page.head.set_location_value(location_list, item)
    elif (location_string == 'body'):
        page.body.set_location_value(location_list, item)
    elif (location_string == 'svg'):
        page.svg.set_location_value(location_list, item)
    elif container is not None: 
        # location is in widget container
        if location_string not in page.widgets:
            raise ServerError("Location of item not found")
        widget = page.widgets[location_string]
        widget.set_in_container(container, location_list, item)
    else:
        raise ServerError("Location of item not found")
    # And save this page copy to the project
    return proj.save_page(page)


def set_item_in_section(project, section_name, schange, location, item):
    "Sets an item in the location"

    proj, section = get_proj_section(project, section_name, schange)

    location_string, container, location_list = location

    # set item in a section
    if location_string == section_name:
        # part is not embedded in a widget
        section.set_location_value(location_list, item)
    elif container is not None: 
        # location is in widget container
        if location_string not in section.widgets:
            raise ServerError("Location of item not found")
        widget = section.widgets[location_string]
        widget.set_in_container(container, location_list, item)
    else:
        raise ServerError("Location of item not found")
    # And save this section copy to the project
    return proj.add_section(section_name, section)



